"""File with the VoidBox class"""

import logging

import FreeCAD
import Part

from ..conversion import cell_definition as Conv
from ..decompose import decom_one as Decom
from ..utils.basic_functions_part1 import is_opposite
from ..utils.boolean_function import BoolSequence
from ..utils.boolean_solids import build_c_table_from_solids, remove_extra_surfaces
from ..utils.functions import GeounedSolid, GeounedSurface

logger = logging.getLogger("general_logger")


class VoidBox:
    def __init__(self, MetaSolids, Enclosure):

        self.Objects = []
        if "BoundBox" in str(Enclosure):
            self.BoundBox = Enclosure
            self.PieceEnclosure = None
        else:
            self.BoundBox = Enclosure.optimalBoundingBox()
            self.PieceEnclosure = Enclosure

        for m in MetaSolids:
            if not m.BoundBox:
                continue
            if m.BoundBox.isValid():
                if self.BoundBox.intersect(m.BoundBox):
                    Obj = self.copy_meta(m)
                    self.remove_extra_comp(Obj, self.BoundBox)
                    self.Objects.append(Obj)
        return

    def split(self, minSize=200):

        dims = [self.BoundBox.XLength, self.BoundBox.YLength, self.BoundBox.ZLength]
        coord = ["X", "Y", "Z"]
        for i in range(2, -1, -1):
            if 0.5 * dims[i] < minSize:
                del dims[i]
                del coord[i]

        if len(dims) == 0:
            return None

        ip = dims.index(max(dims))
        #    print ('dims : {} {}'.format(coord[ip],0.5*dims[ip]))
        if coord[ip] == "X":
            pos = self.BoundBox.XMin + 0.5 * self.BoundBox.XLength
            X1Min = self.BoundBox.XMin
            X1Max = pos
            X2Min = pos
            X2Max = self.BoundBox.XMax
            Y1Min = self.BoundBox.YMin
            Y1Max = self.BoundBox.YMax
            Y2Min = Y1Min
            Y2Max = Y1Max
            Z1Min = self.BoundBox.ZMin
            Z1Max = self.BoundBox.ZMax
            Z2Min = Z1Min
            Z2Max = Z1Max
        elif coord[ip] == "Y":
            pos = self.BoundBox.YMin + 0.5 * self.BoundBox.YLength
            X1Min = self.BoundBox.XMin
            X1Max = self.BoundBox.XMax
            X2Min = X1Min
            X2Max = X1Max
            Y1Min = self.BoundBox.YMin
            Y1Max = pos
            Y2Min = pos
            Y2Max = self.BoundBox.YMax
            Z1Min = self.BoundBox.ZMin
            Z1Max = self.BoundBox.ZMax
            Z2Min = Z1Min
            Z2Max = Z1Max
        else:
            pos = self.BoundBox.ZMin + 0.5 * self.BoundBox.ZLength
            X1Min = self.BoundBox.XMin
            X1Max = self.BoundBox.XMax
            X2Min = X1Min
            X2Max = X1Max

            Y1Min = self.BoundBox.YMin
            Y1Max = self.BoundBox.YMax
            Y2Min = Y1Min
            Y2Max = Y1Max

            Z1Min = self.BoundBox.ZMin
            Z1Max = pos
            Z2Min = pos
            Z2Max = self.BoundBox.ZMax

        VMin1 = FreeCAD.Vector(X1Min, Y1Min, Z1Min)
        VMax1 = FreeCAD.Vector(X1Max, Y1Max, Z1Max)
        VMin2 = FreeCAD.Vector(X2Min, Y2Min, Z2Min)
        VMax2 = FreeCAD.Vector(X2Max, Y2Max, Z2Max)
        box1 = FreeCAD.BoundBox(VMin1, VMax1)
        box2 = FreeCAD.BoundBox(VMin2, VMax2)

        if self.PieceEnclosure is None:
            Space1 = VoidBox(self.Objects, box1)
            Space2 = VoidBox(self.Objects, box2)
            VoidBoxTuple = (Space1, Space2)
        else:
            Space1 = self.piece_enclosure_split(box1)
            Space2 = self.piece_enclosure_split(box2)
            VoidBoxTuple = (Space1, Space2)
            if Space1 is None:
                VoidBoxTuple = (Space2,)
            if Space2 is None:
                VoidBoxTuple = (Space1,)
            if Space1 is None and Space2 is None:
                VoidBoxTuple = ()

        return VoidBoxTuple

    def piece_enclosure_split(self, Box, Tolerance=1.0e-13):
        """This function creates a box-shaped solid with the new limits of given bounding box and
        it is intersected with the piece of nested enclosure to create the new void cell.
        If the limited region does not intersect with the piece, no void cell is created.
        """

        Cube = Part.makeBox(
            Box.XLength,
            Box.YLength,
            Box.ZLength,
            FreeCAD.Vector(Box.XMin, Box.YMin, Box.ZMin),
            FreeCAD.Vector(0, 0, 1),
        )
        dist = Cube.distToShape(self.PieceEnclosure)[0]
        try:
            if abs(dist / Box.DiagonalLength) > Tolerance:
                return None
        except ZeroDivisionError:
            return None
        ShapeObject = Cube.common(self.PieceEnclosure)
        try:
            reldif = (Cube.Volume - ShapeObject.Volume) / Cube.Volume
        except ZeroDivisionError:
            return None
        if abs(reldif) <= Tolerance:
            return VoidBox(self.Objects, Box)
        elif ShapeObject.Solids:
            Solid = ShapeObject.Solids[0]
            return VoidBox(self.Objects, Solid)
        else:
            return None

    def refine(self):
        Cube = Part.makeBox(
            self.BoundBox.XLength,
            self.BoundBox.YLength,
            self.BoundBox.ZLength,
            FreeCAD.Vector(self.BoundBox.XMin, self.BoundBox.YMin, self.BoundBox.ZMin),
            FreeCAD.Vector(0, 0, 1),
        )

        for m in self.Objects:
            self.remove_extra_comp(m, Cube, mode="dist")
        return

    def get_void_complementary(self, Surfaces, options, tolerances, numeric_format, simplify="no"):
        if self.PieceEnclosure is None:
            boxDef = BoolSequence(operator="AND")
            center = self.BoundBox.Center
            bBox = self.BoundBox
            for p in self.get_bound_planes():
                id, exist = Surfaces.add_plane(p, options, tolerances, numeric_format, False)
                if exist:
                    s = Surfaces.get_surface(id)
                    if is_opposite(p.Surf.Axis, s.Surf.Axis):
                        id = -id
                if is_opposite(p.Surf.Axis, p.Surf.Position - center):
                    boxDef.elements.append(id)
                else:
                    boxDef.elements.append(-id)
                enclosure = False
            d = options.enlargeBox

        else:
            UniverseBox = self.PieceEnclosure.BoundBox
            TempPieceEnclosure = GeounedSolid(None, self.PieceEnclosure)
            comsolid, err = Decom.SplitSolid(
                Part.makeCompound(TempPieceEnclosure.Solids),
                UniverseBox,
                options,
                tolerances,
                numeric_format,
            )
            Surfaces.extend(
                Decom.extract_surfaces(
                    comsolid,
                    "All",
                    UniverseBox,
                    options,
                    tolerances,
                    numeric_format,
                    MakeObj=True,
                ),
                options,
                tolerances,
                numeric_format,
            )
            TempPieceEnclosure.update_solids(comsolid.Solids)
            Conv.cellDef(
                TempPieceEnclosure,
                Surfaces,
                UniverseBox,
                options,
                tolerances,
                numeric_format,
            )

            boxDef = TempPieceEnclosure.Definition
            bBox = self.PieceEnclosure.BoundBox
            enclosure = True
            d = max(options.enlargeBox, 2)

        Box = Part.makeBox(
            bBox.XLength + 2 * d,
            bBox.YLength + 2 * d,
            bBox.ZLength + 2 * d,
            FreeCAD.Vector(bBox.XMin - d, bBox.YMin - d, bBox.ZMin - d),
            FreeCAD.Vector(0, 0, 1),
        )

        voidSolidDef = BoolSequence(operator="OR")

        cellIn = []
        for m in self.Objects:
            voidSolidDef.append(m.Definition)
            if m.IsEnclosure:
                continue
            cellIn.append(m.__id__)

        # voidSolidDef.join_operators()

        if not voidSolidDef.elements:
            return (
                boxDef,
                None,
            )  #  Cell to get complementary are null => Void is only box definition

        # join all basic solids into one big meta Object
        # CAD solid representation is not needed because
        # here we are working with surfaces and void box

        complementary = BoolSequence(operator="AND")
        complementary.append(boxDef)
        if simplify != "no":
            surfList = voidSolidDef.get_surfaces_numbers()

            if enclosure:
                surfList.update(boxDef.get_surfaces_numbers())
            else:
                for s in boxDef.elements:
                    val = s > 0
                    voidSolidDef.substitute(abs(s), val)
                voidSolidDef.clean()
            if type(voidSolidDef.elements) is bool:
                res = voidSolidDef.elements
            else:
                res = None

            if enclosure or res is None:
                surfaceDict = {}
                for i in surfList:
                    surfaceDict[i] = Surfaces.get_surface(i)
                CTable = build_c_table_from_solids(Box, surfaceDict, simplify, options=options)
            else:
                if res is True:
                    return None, None
                else:
                    return boxDef, None

            newTemp = BoolSequence(operator="OR")

            if voidSolidDef.level == 0:
                if len(voidSolidDef.elements) == 1:
                    voidSolidDef.operator = "AND"
                cellVoid = BoolSequence(operator="OR")
                cellVoid.append(voidSolidDef)
                voidSolidDef = cellVoid

            for solDef in voidSolidDef.elements:
                newSolid = remove_extra_surfaces(solDef, CTable)
                if type(newSolid.elements) is not bool:
                    newTemp.append(newSolid)
                elif newSolid.elements is True:
                    return None, None

            voidSolidDef = newTemp

        else:
            if voidSolidDef.level == 0:
                if len(voidSolidDef.elements) == 1:
                    voidSolidDef.operator = "AND"
                cellVoid = BoolSequence(operator="OR")
                cellVoid.append(voidSolidDef)
                voidSolidDef = cellVoid

        if voidSolidDef.elements is True:
            return None, None
        elif voidSolidDef.elements is False or voidSolidDef.elements == []:
            return boxDef, None

        if voidSolidDef.level == 0:
            compSeq = voidSolidDef.get_complementary()
        else:

            if voidSolidDef.level == 1 and voidSolidDef.operator == "AND":
                compSeq = BoolSequence(operator="OR")
            else:
                compSeq = BoolSequence(operator="AND")

            for comp in voidSolidDef.elements:
                if simplify == "no":
                    comp.check()
                    if type(comp.elements) is bool:
                        chk = comp.elements
                    else:
                        chk = None

                    # solid in cover full Void cell volume  => Void cell doesn't exist
                    if chk is True:
                        logger.warning("void Cell should not exist")
                        return None, None

                    # solid cell is not in void cell Void cell volume  => doesn't contribute to void definition
                    elif chk is False:
                        continue

                pmoc = comp.get_complementary()
                compSeq.append(pmoc)

        if simplify == "full":
            if enclosure:
                complementary.append(compSeq)
                complementary.simplify(CTable)
            else:
                compSeq.simplify(CTable)
                complementary.append(compSeq)
        else:
            compSeq.simplify(None)
            complementary.simplify(None)
            complementary.append(compSeq)

        complementary.clean()
        complementary.level_update()

        if type(complementary.elements) is bool:
            return None, None
        else:
            return complementary, cellIn

    def get_numbers(self):
        ns = 0
        nb = 0

        for m in self.Objects:
            ns += len(m.Surfaces)
            nb += len(m.Definition.elements) if m.Definition.level > 0 else 1

        return ns, nb

    def get_bound_planes(self):
        Xmid = 0.5 * (self.BoundBox.XMin + self.BoundBox.XMax)
        Ymid = 0.5 * (self.BoundBox.YMin + self.BoundBox.YMax)
        Zmid = 0.5 * (self.BoundBox.ZMin + self.BoundBox.ZMax)
        LX = self.BoundBox.ZMin + self.BoundBox.XLength
        LY = self.BoundBox.ZMin + self.BoundBox.YLength
        LZ = self.BoundBox.ZMin + self.BoundBox.ZLength
        PXMin = GeounedSurface(
            (
                "Plane",
                (
                    FreeCAD.Vector(self.BoundBox.XMin, Ymid, Zmid),
                    FreeCAD.Vector(1, 0, 0),
                    LY,
                    LZ,
                ),
            ),
            self.BoundBox,
        )
        PXMax = GeounedSurface(
            (
                "Plane",
                (
                    FreeCAD.Vector(self.BoundBox.XMax, Ymid, Zmid),
                    FreeCAD.Vector(-1, 0, 0),
                    LY,
                    LZ,
                ),
            ),
            self.BoundBox,
        )
        PYMin = GeounedSurface(
            (
                "Plane",
                (
                    FreeCAD.Vector(Xmid, self.BoundBox.YMin, Zmid),
                    FreeCAD.Vector(0, 1, 0),
                    LZ,
                    LX,
                ),
            ),
            self.BoundBox,
        )
        PYMax = GeounedSurface(
            (
                "Plane",
                (
                    FreeCAD.Vector(Xmid, self.BoundBox.YMax, Zmid),
                    FreeCAD.Vector(0, -1, 0),
                    LZ,
                    LX,
                ),
            ),
            self.BoundBox,
        )
        PZMin = GeounedSurface(
            (
                "Plane",
                (
                    FreeCAD.Vector(Xmid, Ymid, self.BoundBox.ZMin),
                    FreeCAD.Vector(0, 0, 1),
                    LX,
                    LY,
                ),
            ),
            self.BoundBox,
        )
        PZMax = GeounedSurface(
            (
                "Plane",
                (
                    FreeCAD.Vector(Xmid, Ymid, self.BoundBox.ZMax),
                    FreeCAD.Vector(0, 0, -1),
                    LX,
                    LY,
                ),
            ),
            self.BoundBox,
        )

        return (PXMin, PXMax, PYMin, PYMax, PZMin, PZMax)

    def remove_extra_comp(self, Obj, Box, mode="box"):
        reducedSol = []
        reducedDef = BoolSequence(operator="OR")
        if not Obj.Solids:
            return
        # Compare Solid BoundBox (here Box is BoundBox Object)
        if mode == "box":
            for i, sol in enumerate(Obj.Solids):
                if sol.BoundBox.isValid():
                    if Box.intersect(sol.BoundBox):
                        reducedSol.append(sol)
                        reducedDef.append(Obj.Definition.elements[i])

        # Compare solid using distToshape (here Box is a Solid Cube object)
        else:
            for i, sol in enumerate(Obj.Solids):
                dist = Box.distToShape(sol)[0]
                if dist == 0:
                    reducedSol.append(sol)
                    reducedDef.append(Obj.Definition.elements[i])

        if len(reducedSol) < len(Obj.Solids):
            Obj.update_solids(reducedSol)
            Obj.set_definition(reducedDef)
        return

    def copy_meta(self, m):
        solidsCopy = m.Solids[:]
        facesCopy = m.Faces[:]
        Meta = GeounedSolid(m.__id__, solidsCopy)
        Meta.set_definition(m.Definition.copy())
        Meta.set_faces(facesCopy)
        if m.IsEnclosure:
            Meta.IsEnclosure = True
            Meta.EnclosureID = m.EnclosureID
        return Meta



import os
import subprocess
import sys
from pathlib import Path

sys.path.append(str(Path(__file__).parents[1] / "src"))
sys.path.append("/usr/lib64/freecad/lib64/")

from geouned import CadToCsg


def setInput(inName, inpDir, outDir):

    if inName.endswith(".step"):
        filename = inName[0:-5]
    elif inName.endswith(".stp"):
        filename = inName[0:-4]
    else:
        filename = inName

    if inpDir == "":
        inpDir = "."
    if outDir == "":
        outDir = "."

    inName = f"{inpDir}/{inName}"
    outName = f"{outDir}/{filename}"

    template = """[Files]
title    = Input Test
stepFile = {}
geometryName = {}

[Parameters]
compSolids = False
volCARD    = False
volSDEF    = True
voidGen    = True
dummyMat    = True
minVoidSize =  100
cellSummaryFile = False
cellCommentFile = False
debug       = False
simplify   = full

[Options]
forceCylinder = False
splitTolerance = 0
newSplitPlane = True
nPlaneReverse = 0
""".format(
        inName, outName
    )

    with open(file="config.ini", mode="w", encoding="utf-8") as outfile:
        outfile.write(template)


def getInputList(folder, ext=None):
    filenameList = []
    if ext is None:
        return os.listdir(folder)

    elif type(ext) is str:
        if ext[0] != ".":
            ext = "." + ext
        for f in os.listdir(folder):
            if f.endswith(ext):
                filenameList.append(f)

    elif type(ext) is list or type(ext) is tuple:
        filenameList = []
        extList = []
        for e in ext:
            if e[0] != ".":
                e = "." + e
            extList.append(e)

        for f in os.listdir(folder):
            for e in extList:
                if f.endswith(e):
                    filenameList.append(f)

    return filenameList


def runMCNP(path, inpFile):
    code = "/opt/mcnp5/last/bin/mcnp5"
    xsdir = "/opt/mcnp.data/ascii/xsdir"
    inp = inpFile
    out = inpFile[0:-1] + "o"
    mctal = inpFile[0:-1] + "m"
    cmd = "cd {} && {} i={} o={} mctal={} xsdir={}".format(path, code, inp, out, mctal, xsdir)
    os.system(cmd)


def clean(folder):
    filename = "{}/runtpe".format(folder)
    if os.path.isfile(filename):
        os.remove(filename)


def cleanDir(folder):
    for file in getInputList(folder, (".o", ".m")):
        filename = "{}/{}".format(folder, file)
        os.remove(filename)


def checkLost(outp):
    cmd = "grep 'particles got lost.' {}".format(outp)
    stdout = subprocess.getoutput(cmd)
    lineout = stdout.split("\n")
    lost = 0
    for l in lineout:
        if l == "" or "terminated" in l:
            continue
        lost = int(l[0 : l.index("particles")])
    return lost


def getMctalValues(mctal):
    file = open(mctal, "r")
    while True:
        line = file.readline()
        if "vals" in line:
            break

    rawData = []
    line = file.readline()
    while "tfc" not in line:
        rawData.extend(line.split())
        line = file.readline()
    file.close()

    values = []
    for i in range(int(len(rawData) / 2)):
        val = float(rawData[2 * i])
        err = float(rawData[2 * i + 1])
        values.append((val, err))
    return values


def printResults(f, res, lost):

    line = "{} :\n".format(f)
    if lost != 0:
        line += "{} particles got lost.\n".format(lost)

    for i, vals in enumerate(res):
        line += "  {:<3d} : {:8.5f}  {:7.5f}\n".format(i + 1, vals[0], vals[1])
    print(line)


def mkGEOInp(inpDir, outDir):
    for f in getInputList(inpDir, ("stp", "step")):
        setInput(f, inpDir, outDir)
        GEO = CadToCsg()
        GEO.set_configuration(inifile)
        GEO.Start()
        del GEO


def processMCNPfolder(outDir):
    cleanDir(outDir)
    for f in getInputList(outDir, ".mcnp"):
        os.rename(f"{outDir}/{f}", f"{outDir}/{f[:-4]}i")

    for f in getInputList(outDir, ".i"):
        runMCNP(outDir, f)
        clean(outDir)


def postProcess(folder):
    for fmct in getInputList(folder, ".m"):
        fout = fmct[:-1] + "o"
        mctal = "{}/{}".format(folder, fmct)
        outp = "{}/{}".format(folder, fout)
        res = getMctalValues(mctal)
        lost = checkLost(outp)
        printResults(fmct[:-1] + "i", res, lost)


# ****************************************************

misInp = "inputSTEP/Misc"
misOut = "outMCNP/Misc"
cylInp = "inputSTEP/DoubleCylinder"
cylOut = "outMCNP/DoubleCylinder"
torInp = "inputSTEP/Torus"
torOut = "outMCNP/Torus"
lrgInp = "inputSTEP/large"
lrgOut = "outMCNP/large"
ms0Inp = "inputSTEP/"
ms0Out = "outMCNP/"

inifile = "config.ini"
folder = {
    "misc": (misInp, misOut),
    "cyl": (cylInp, cylOut),
    "torus": (torInp, torOut),
    "large": (lrgInp, lrgOut),
    "misc0": (ms0Inp, ms0Out),
}

test = "all"

if test == "all":
    for inpDir, outDir in folder.values():
        mkGEOInp(inpDir, outDir)

    for inpDir, outDir in folder.values():
        processMCNPfolder(outDir)

    for inpDir, outDir in folder.values():
        postProcess(outDir)

else:
    inpDir, outDir = folder[test]
    mkGEOInp(inpDir, outDir)
    processMCNPfolder(outDir)
    postProcess(outDir)



import logging

formatter = logging.Formatter("%(asctime)s :: %(levelname)s :: %(funcName)s :: %(lineno)d :: %(message)s")


def setup_logger(name, log_file, level=logging.DEBUG):
    """To setup as many loggers as you want"""

    handler = logging.FileHandler(log_file)
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger



def assignEnclosure(MetaList, NestedEnclosure):
    """Assign to all solids the Enclosures ID it belongs to"""

    for m in MetaList:
        if m.IsEnclosure:
            continue
        dep = set()
        outside = True
        for lev, Level in enumerate(reversed(NestedEnclosure)):
            for encl in Level:
                inter = m.check_intersection(encl.CADSolid)
                # CAD solid intersect level i enclosure -> end of nested loops
                # Current loop continue because solid can intersect various enclosure
                if inter == 0:
                    dep.add(encl.EnclosureID)
                    m.ParentEnclosureID = encl.EnclosureID
                    if lev == len(NestedEnclosure) - 1:
                        dep.add(-1)
                        m.ParentEnclosureID = -1

                # CAD solid inside level i enclosure -> end of current loop go up i-1 level
                elif inter == -1:
                    dep.add(encl.EnclosureID)
                    m.ParentEnclosureID = encl.EnclosureID
                    outside = False
                    break

            # stop loop if CAD solid outside all level i enclosure or intersecting one
            if not outside:
                m.enclosure_list = dep
                break

        if not dep:
            dep.add(-1)
            m.ParentEnclosureID = -1
        if outside:
            m.enclosure_list = dep
    return


def select_solids(MetaList, LowLevelEnclosure, Enclosure):
    """This function selects the CAD solids and nested enclosures
    that can intersect with the current nested enclosure.
    For intermediate Levels selected solids are those intersecting i and i+1 level enclosure,
    or fully inside level i enclosure and fully outside i+1 enclosure"""

    newMetaList = LowLevelEnclosure[:]

    if "BoundBox" in str(Enclosure):
        enclID = -1
    else:
        enclID = Enclosure.EnclosureID

    for m in MetaList:
        if m.IsEnclosure:
            continue
        if enclID in m.enclosure_list:
            newMetaList.append(m)

    return newMetaList


def update_void_list(offset, voidList, NestedEnclosure, sort_enclosure):

    newVoidList = []
    if NestedEnclosure:
        update_comment = True
    else:
        update_comment = False

    icount = offset + 1
    voids = voidList[0]
    for m in voids:
        m.__id__ = icount
        if update_comment and not sort_enclosure:
            m.Comments = m.Comments + "\nLevel 0 void enclosure"
        icount += 1
        m.ParentEnclosureID = -1
        m.EnclosureID = 0
        newVoidList.append(m)

    ivoid = 1
    for Level in NestedEnclosure:
        for encl in Level:
            if encl.CellType == "envelope":
                continue
            for m in voidList[ivoid]:
                m.__id__ = icount
                m.ParentEnclosureID = encl.ParentEnclosureID
                m.EnclosureID = encl.EnclosureID
                if sort_enclosure:
                    m.Comments = m.Comments + f"\n{encl.Comments}"
                else:
                    m.Comments = m.Comments + f"\nVoid Enclosure #{encl.EnclosureID}"
                icount += 1
                newVoidList.append(m)
            ivoid += 1

    # graveyard Cell

    Graveyard = voidList[-1]
    for m in Graveyard:
        m.__id__ = icount
        m.ParentEnclosureID = -1
        m.EnclosureID = 0
        newVoidList.append(m)
        icount += 1

    return newVoidList



import math

import FreeCAD
import numpy as np
import Part

from .buildSolidCell import BuildSolid
from .remh import Cline
from .Utils.booleanFunction import BoolSequence, outer_terms


class CadCell:
    def __init__(self, stringCell=None):

        if not stringCell:
            self.surfaces = {}
            self.surfaceList = []
            self.shape = None
            # self.likeCell = None
            self.definition = None
            self.name = 0
            # self.TRCL     = None  # cell transformacion "like-but" cells
            self.TRFL = None  # Universe transformation in fill Universe
            self.U = -1  # Cell Universe number
            self.FILL = 0  # Fill Universe number
            self.MAT = 0  # material number
            self.CurrentTR = None
            self.level = None
            self.__defTerms__ = None
            self.__operator__ = None
        else:
            self.surfaces = None
            self.shape = None
            self.name = stringCell.name
            self.TRFL = stringCell.TR  # Universe transformation in fill Universe
            self.U = stringCell.U  # Cell Universe number
            self.FILL = stringCell.FILL  # Fill Universe number
            self.MAT = stringCell.MAT  # material number
            self.CurrentTR = self.TRFL
            self.level = None

            self.__defTerms__ = None
            self.__operator__ = None
            self.__setDefinition__(stringCell)

    def copy(self):
        cpCell = CadCell()
        cpCell.surfaceList = self.surfaceList[:]
        cpCell.surfaces = {}
        for name, s in self.surfaces.items():
            cpCell.surfaces[name] = s.copy()

        if type(self.definition) is Cline:
            cpCell.definition = Cline(self.definition.str)

        elif type(self.definition) is BoolSequence:
            cpCell.definition = self.definition.copy()

        cpCell.name = self.name
        cpCell.TRFL = self.TRFL
        cpCell.U = self.U
        cpCell.FILL = self.FILL
        cpCell.MAT = self.MAT
        cpCell.level = self.level

        if self.CurrentTR is not None:
            cpCell.CurrentTR = self.CurrentTR.submatrix(4)

        if self.shape is not None:
            cpCell.shape = self.shape.copy()

        return cpCell

    def getSubCell(self, seq):

        subCell = self.copy()
        subCell.definition = seq.copy()
        subCell.shape = None
        subCell.surfaceList = subCell.definition.get_surfaces_numbers()
        for s in tuple(subCell.surfaces.keys()):
            if s not in subCell.surfaceList:
                del subCell.surfaces[s]

        return subCell

    # not used
    #
    #    def split(self,nparts=2):
    #
    #        if nparts == 1:
    #            return (self,None)
    #        terms,operador = self.getOuterTerms()
    #        nelemts = int(len(terms)/nparts)
    #        subDefList = []
    #
    #        if operador == 'AND':
    #          for i in range(nparts-1):
    #             newdef = ') ('.join(terms[i*nelemts:(i+1)*nelemts])
    #             newdef = '({})'.format(newdef)
    #             subDefList.append(newdef)
    #          newdef = ') ('.join(terms[(nparts-1)*nelemts:])
    #          newdef = '({})'.format(newdef)
    #          subDefList.append(newdef)
    #
    #        else:
    #          for i in range(nparts-1):
    #             newdef = '):('.join(terms[i*nelemts:(i+1)*nelemts])
    #             newdef = '({})'.format(newdef)
    #             subDefList.append(newdef)
    #          newdef = '):('.join(terms[(nparts-1)*nelemts:])
    #          newdef = '({})'.format(newdef)
    #          subDefList.append(newdef)
    #
    #
    #        subCellList=[]
    #        for df in subDefList:
    #           subCell = self.copy()
    #           subCell.definition= Cline(df)
    #           subCell.shape = None
    #           subCell.surfaceList  = subCell.definition.get_surfaces_numbers()
    #           for s in tuple(subCell.surfaces.keys()) :
    #               if s not in subCell.surfaceList: del(subCell.surfaces[s])
    #
    #           subCellList.append(subCell)
    #
    #        return subCellList,operador

    def getOuterTerms(self):
        if not self.__defTerms__:
            self.__defTerms__, self.__operator__ = outer_terms(self.definition.str)
        return self.__defTerms__, self.__operator__

    def makeBox(self, boundBox):
        box_origin = FreeCAD.Vector(boundBox.XMin, boundBox.YMin, boundBox.ZMin)
        return Part.makeBox(boundBox.XLength, boundBox.YLength, boundBox.ZLength, box_origin)

    def buildShape(self, boundBox, force=False, surfTR=None, simplify=False, fuse=False):

        if self.shape is not None and not force:
            return
        if surfTR:
            self.transformSurfaces(surfTR)

        cutShape = BuildSolid(self, boundBox, simplify=simplify)

        # TODO consider making this step conditional on fuse
        # if fuse or True:
        #     self.shape = FuseSolid(cutShape)
        # else:
        #     self.shape = Part.makeCompound(cutShape)
        self.shape = FuseSolid(cutShape)

    def buildSurfaceShape(self, boundBox):
        for s in self.surfaces.values():
            s.buildShape(boundBox)

    def transformSolid(self, matrix, reverse=False):
        if not self.shape:
            return
        if reverse:
            self.shape = self.shape.transformGeometry(matrix.inverse())
        else:
            self.shape = self.shape.transformGeometry(matrix)

    def transformSurfaces(self, matrix):
        for s in self.surfaces.values():
            s.transform(matrix)

    def setSurfaces(self, Surfaces):
        if self.surfaces is not None:
            return
        self.surfaces = {}
        for s in self.surfaceList:
            self.surfaces[s] = Surfaces[s]

    def cleanUndefined(self):
        undefined = []
        for s in self.definition.get_surfaces_numbers():
            if self.surfaces[s].params is None:
                undefined.append(s)
        if undefined:
            self.definition.removeSurface(undefined)

        for s in undefined:
            del self.surfaces[s]

    def __setDefinition__(self, stringCell):

        self.definition = stringCell.geom
        self.definition.remove_comments(full=True)
        self.definition.remove_cr()
        self.definition.remove_multispace()
        self.definition.remove_redundant()
        self.surfaceList = self.definition.get_surfaces_numbers()


class Plane:
    def __init__(self, Id, params, tr=None):
        self.type = "plane"
        self.id = Id
        self.shape = None
        self.params = params
        if tr:
            self.transform(tr)

    def __str__(self):
        return f"plane : {self.id}\nParameters : {self.params}"

    def copy(self):
        return Plane(self.id, self.params)

    def transform(self, matrix):
        v, d = self.params
        p = d * v  # vector p is d*plane normal
        v = matrix.submatrix(3).multVec(v)
        v.normalize()
        d = matrix.multVec(p) * v
        self.params = (v, d)

    def buildShape(self, boundBox):
        normal, p0 = self.params
        Box = FreeCAD.BoundBox(boundBox)
        Box.enlarge(10)

        pointEdge = []
        for i in range(12):
            edge = Box.getEdge(i)
            p1 = normal.dot(edge[0])
            p2 = normal.dot(edge[1])
            d0 = p0 - p1
            d1 = p2 - p1
            if d1 != 0:
                a = d0 / d1
                if a >= 0 and a <= 1:
                    pointEdge.append(edge[0] + a * (edge[1] - edge[0]))

        if len(pointEdge) == 0:
            return
        s = FreeCAD.Vector((0, 0, 0))
        for v in pointEdge:
            s = s + v
        s = s / len(pointEdge)

        vtxvec = []
        for v in pointEdge:
            vtxvec.append(v - s)

        X0 = vtxvec[0]
        Y0 = normal.cross(X0)

        orden = []
        for i, v in enumerate(vtxvec):
            phi = np.arctan2(v.dot(Y0), v.dot(X0))
            orden.append((phi, i))
        orden.sort()

        self.shape = Part.Face(Part.makePolygon([pointEdge[p[1]] for p in orden], True))


class Sphere:
    def __init__(self, Id, params, tr=None):
        self.type = "sphere"
        self.id = Id
        self.shape = None
        self.params = params
        if tr:
            self.transform(tr)

    def copy(self):
        return Sphere(self.id, self.params)

    def transform(self, matrix):
        p, R = self.params
        p = matrix.multVec(p)
        self.params = (p, R)

    def buildShape(self, boundBox):
        origin, R = self.params
        self.shape = Part.makeSphere(R, origin)


class Cylinder:
    def __init__(self, Id, params, tr=None, truncated=False):
        self.type = "cylinder"
        self.id = Id
        self.shape = None
        self.params = params
        self.truncated = truncated
        if tr:
            self.transform(tr)

    def copy(self):
        return Cylinder(self.id, self.params, truncated=self.truncated)

    def transform(self, matrix):
        p, v, R = self.params
        v = matrix.submatrix(3).multVec(v)
        p = matrix.multVec(p)
        self.params = (p, v, R)

    def buildShape(self, boundBox):

        p, vec, r = self.params

        if not self.truncated:
            dmin = vec.dot(boundBox.getPoint(0) - p)
            dmax = dmin
            for i in range(1, 8):
                d = vec.dot(boundBox.getPoint(i) - p)
                dmin = min(d, dmin)
                dmax = max(d, dmax)

            height = dmax - dmin
            dmin -= 0.1 * height
            dmax += 0.1 * height
            height = dmax - dmin

            point = p + dmin * vec
            self.shape = Part.makeCylinder(r, height, point, vec, 360)
            # self.shape = makeCylinder2( r,height,point,vec)
        else:
            self.shape = Part.makeCylinder(r, vec.Length, p, vec, 360)
            # self.shape = Part.makeCylinder2( r,vec.Length,p,vec)

        return


class Cone:
    def __init__(self, Id, params, tr=None, truncated=False):
        self.type = "cone"
        self.id = Id
        self.shape = None
        self.params = params
        self.truncated = truncated
        if tr:
            self.transform(tr)

    def copy(self):
        return Cone(self.id, self.params, truncated=self.truncated)

    def transform(self, matrix):
        if not self.truncated:
            p, v, t, dbl = self.params
            v = matrix.submatrix(3).multVec(v)
            p = matrix.multVec(p)
            self.params = (p, v, t, dbl)
        else:
            p, v, r1, r2 = self.params
            v = matrix.submatrix(3).multVec(v)
            p = matrix.multVec(p)
            self.params = (p, v, r1, r2)

    def buildShape(self, boundBox):
        if not self.truncated:
            apex, axis, t, dblsht = self.params

            dmin = axis.dot(boundBox.getPoint(0) - apex)
            dmax = dmin
            for i in range(1, 8):
                d = axis.dot(boundBox.getPoint(i) - apex)
                dmin = min(d, dmin)
                dmax = max(d, dmax)

            length = max(abs(dmin), abs(dmax))
            R = length * t
            OneSheetCone = Part.makeCone(0, R, length, apex, axis, 360)
            if not dblsht:
                self.shape = OneSheetCone
            else:
                OtherSheet = Part.makeCone(0, R, length, apex, -axis, 360)
                DoubleSheetCone = OneSheetCone.fuse([OtherSheet])
                DoubleSheetCone.removeSplitter()
                self.shape = DoubleSheetCone
        else:
            center, axis, r1, r2 = self.params
            self.shape = Part.makeCone(r1, r2, axis.Length, center, axis, 360)


class EllipticCone:
    def __init__(self, Id, params, tr=None):
        self.type = "cone_elliptic"
        self.id = Id
        self.shape = None
        self.params = params
        if tr:
            self.transform(tr)

    def copy(self):
        return EllipticCone(self.id, self.params)

    def transform(self, matrix):
        p, v, ra, radii, raxes, dbl = self.params
        v = matrix.submatrix(3).multVec(v)
        raxes[0] = matrix.submatrix(3).multVec(raxes[0])
        raxes[1] = matrix.submatrix(3).multVec(raxes[1])
        p = matrix.multVec(p)
        self.params = (p, v, ra, radii, raxes, dbl)

    def buildShape(self, boundBox):
        apex, axis, ra, radii, raxes, dblsht = self.params

        dmin = axis.dot(boundBox.getPoint(0) - apex)
        dmax = dmin
        for i in range(1, 8):
            d = axis.dot(boundBox.getPoint(i) - apex)
            dmin = min(d, dmin)
            dmax = max(d, dmax)

        length = max(abs(dmin), abs(dmax))
        OneSheetCone = makeEllipticCone(apex, axis, ra, radii, raxes, length)
        if not dblsht:
            self.shape = OneSheetCone
        else:
            OtherSheet = makeEllipticCone(apex, -axis, ra, radii, raxes, length)
            DoubleSheetCone = OneSheetCone.fuse([OtherSheet])
            DoubleSheetCone.removeSplitter()
            self.shape = DoubleSheetCone


class Hyperboloid:
    def __init__(self, Id, params, tr=None):
        self.type = "hyperboloid"
        self.id = Id
        self.shape = None
        self.params = params
        if tr:
            self.transform(tr)

    def copy(self):
        return Hyperboloid(self.id, self.params)

    def transform(self, matrix):
        p, v, radii, raxes, onesht = self.params
        v = matrix.submatrix(3).multVec(v)
        raxes[0] = matrix.submatrix(3).multVec(raxes[0])
        raxes[1] = matrix.submatrix(3).multVec(raxes[1])
        p = matrix.multVec(p)
        self.params = (p, v, radii, raxes, onesht)

    def buildShape(self, boundBox):
        center, axis, radii, rAxes, onesht = self.params

        dmin = axis.dot(boundBox.getPoint(0) - center)
        dmax = dmin
        for i in range(1, 8):
            d = axis.dot(boundBox.getPoint(i) - center)
            dmin = min(d, dmin)
            dmax = max(d, dmax)

        length = max(abs(dmin), abs(dmax))
        self.shape = makeHyperboloid(center, radii, rAxes, axis, onesht, length)


class Ellipsoid:
    def __init__(self, Id, params, tr=None):
        self.type = "ellipsoid"
        self.id = Id
        self.shape = None
        self.params = params
        if tr:
            self.transform(tr)

    def copy(self):
        return Ellipsoid(self.id, self.params)

    def transform(self, matrix):
        p, v, radii, raxes = self.params
        v = matrix.submatrix(3).multVec(v)
        raxes[0] = matrix.submatrix(3).multVec(raxes[0])
        raxes[1] = matrix.submatrix(3).multVec(raxes[1])
        p = matrix.multVec(p)
        self.params = (p, v, radii, raxes)

    def buildShape(self, boundBox):
        center, axis, radii, rAxes = self.params
        self.shape = makeEllipsoid(center, radii, rAxes, axis)


class EllipticCylinder:
    def __init__(self, Id, params, tr=None, truncated=False):
        self.type = "cylinder_elliptic"
        self.id = Id
        self.shape = None
        self.params = params
        self.truncated = truncated
        if tr:
            self.transform(tr)

    def copy(self):
        return EllipticCylinder(self.id, self.params, truncated=self.truncated)

    def transform(self, matrix):
        p, v, radii, raxes = self.params
        v = matrix.submatrix(3).multVec(v)
        raxes[0] = matrix.submatrix(3).multVec(raxes[0])
        raxes[1] = matrix.submatrix(3).multVec(raxes[1])
        p = matrix.multVec(p)
        self.params = (p, v, radii, raxes)

    def buildShape(self, boundBox):
        center, axis, radii, rAxes = self.params
        if not self.truncated:
            dmin = axis.dot(boundBox.getPoint(0) - center)
            dmax = dmin
            for i in range(1, 8):
                d = axis.dot(boundBox.getPoint(i) - center)
                dmin = min(d, dmin)
                dmax = max(d, dmax)

            height = dmax - dmin
            dmin -= 0.1 * height
            dmax += 0.1 * height
            height = dmax - dmin
            point = center + dmin * axis

            self.shape = makeEllipticCylinder(point, radii, rAxes, axis, height)
        else:
            height = axis.Length
            self.shape = makeEllipticCylinder(center, radii, rAxes, axis / height, height)


class HyperbolicCylinder:
    def __init__(self, Id, params, tr=None):
        self.type = "cylinder_hyperbolic"
        self.id = Id
        self.shape = None
        self.params = params
        if tr:
            self.transform(tr)

    def copy(self):
        return HyperbolicCylinder(self.id, self.params)

    def transform(self, matrix):
        p, v, radii, raxes = self.params
        v = matrix.submatrix(3).multVec(v)
        raxes[0] = matrix.submatrix(3).multVec(raxes[0])
        raxes[1] = matrix.submatrix(3).multVec(raxes[1])
        p = matrix.multVec(p)
        self.params = (p, v, radii, raxes)

    def buildShape(self, boundBox):
        center, axis, radii, rAxes = self.params

        dmin = axis.dot(boundBox.getPoint(0) - center)
        dmax = dmin
        for i in range(1, 8):
            d = axis.dot(boundBox.getPoint(i) - center)
            dmin = min(d, dmin)
            dmax = max(d, dmax)

        height = dmax - dmin
        dmin -= 0.1 * height
        dmax += 0.1 * height
        height = dmax - dmin
        point = center + dmin * axis

        self.shape = makeHyperbolicCylinder(point, radii, rAxes, axis, height)


class Paraboloid:
    def __init__(self, Id, params, tr=None):
        self.type = "paraboloid"
        self.id = Id
        self.shape = None
        self.params = params
        if tr:
            self.transform(tr)

    def copy(self):
        return Paraboloid(self.id, self.params)

    def transform(self, matrix):
        p, v, focal = self.params
        v = matrix.submatrix(3).multVec(v)
        p = matrix.multVec(p)
        self.params = (p, v, focal)

    def buildShape(self, boundBox):
        center, axis, focal = self.params

        dmin = axis.dot(boundBox.getPoint(0) - center)
        dmax = dmin
        for i in range(1, 8):
            d = axis.dot(boundBox.getPoint(i) - center)
            dmin = min(d, dmin)
            dmax = max(d, dmax)

        length = max(abs(dmin), abs(dmax))
        self.shape = makeParaboloid(center, axis, focal, length)


class Torus:
    def __init__(self, Id, params, tr=None):
        self.type = "torus"
        self.id = Id
        self.shape = None
        self.params = params
        if tr:
            self.transform(tr)

    def copy(self):
        return Torus(self.id, self.params)

    def transform(self, matrix):
        p, v, Ra, Rb, Rc = self.params
        v = matrix.submatrix(3).multVec(v)
        p = matrix.multVec(p)
        self.params = (p, v, Ra, Rb, Rc)

    def buildShape(self, boundBox):
        center, axis, Ra, Rb, Rc = self.params  # Ra distance from torus axis; R radius of toroidal-cylinder
        if (abs(Rb - Rc) < 1e-5) and Ra > 0:
            self.shape = Part.makeTorus(Ra, Rb, center, axis)  # FreeCAD circular Torus
        else:
            self.shape = makeEllipticTorus(Ra, Rb, Rc, center, axis)  # Home made elliptic Torus


class Box:
    def __init__(self, Id, params, tr=None):
        self.type = "box"
        self.id = Id
        self.shape = None
        self.params = params
        if tr:
            self.transform(tr)

    def copy(self):
        return Box(self.id, self.params)

    def transform(self, matrix):
        p, v1, v2, v3 = self.params
        p = matrix.multVec(p)
        v1 = matrix.multVec(v1)
        v2 = matrix.multVec(v2)
        v3 = matrix.multVec(v3)
        self.params = (p, v1, v2, v3)

    def buildShape(self, boundBox):
        p, v1, v2, v3 = self.params
        a1 = FreeCAD.Vector(v1)
        a2 = FreeCAD.Vector(v2)
        a3 = FreeCAD.Vector(v3)
        a1.normalize()
        a2.normalize()
        a3.normalize()

        m = FreeCAD.Matrix(
            a1.x,
            a2.x,
            a3.x,
            p.x,
            a1.y,
            a2.y,
            a3.y,
            p.y,
            a1.z,
            a2.z,
            a3.z,
            p.z,
            0,
            0,
            0,
            1,
        )
        box = Part.makeBox(v1.Length, v2.Length, v3.Length)
        self.shape = box.transformGeometry(m)


class Undefined:
    def __init__(self, Id):
        self.type = "Undefined"
        self.id = Id
        self.shape = None
        self.params = None

    def copy(self):
        return Undefined(self.id)

    def buildShape(self, boundBox):
        return

    def transform(self, matrix):
        return


def FuseSolid(parts):
    if (len(parts)) <= 1:
        if parts:
            solid = parts[0]
        else:
            return None
    else:
        try:
            fused = parts[0].fuse(parts[1:])
        except:
            fused = None

        if fused is not None:
            try:
                refinedfused = fused.removeSplitter()
            except:
                refinedfused = fused

            if refinedfused.isValid():
                solid = refinedfused
            else:
                if fused.isValid():
                    solid = fused
                else:
                    solid = Part.makeCompound(parts)
        else:
            solid = Part.makeCompound(parts)

    if solid.Volume < 0:
        solid.reverse()
    return solid


def makeHyperboloid(center, radii, rAxes, axis, onesht, length):

    S1 = center + rAxes[1] * radii[1]  # major axis
    S2 = center + rAxes[0] * radii[0]  # minor axis
    hyperbola = Part.Hyperbola(S1, S2, center)

    Y = length
    X = radii[1] * math.sqrt((Y / radii[0]) ** 2 + 1)
    point = center + X * radii[1] + Y * radii[0]  # point in taken as length is always counted on minor axis
    parameter = abs(hyperbola.parameter(point))

    if onesht:
        shape = hyperbola.toBSpline(-parameter, parameter).toShape(-parameter, parameter)
        hyperFace = shape.revolve(center, axis, 360)

        StartPoint = hyperFace.Surface.BasisCurve.StartPoint - center
        EndPoint = hyperFace.Surface.BasisCurve.EndPoint - center

        rad1 = StartPoint.dot(hyperbola.XAxis)
        hgt1 = StartPoint.dot(hyperbola.YAxis)
        cc1 = center + hyperbola.YAxis * hgt1
        circle1 = Part.Circle(cc1, -hyperbola.YAxis, rad1).toShape()
        cFace1 = Part.makeFace(circle1, "Part::FaceMakerSimple")

        rad2 = EndPoint.dot(hyperbola.XAxis)
        hgt2 = EndPoint.dot(hyperbola.YAxis)
        cc2 = center + hyperbola.YAxis * hgt2
        circle2 = Part.Circle(cc2, hyperbola.YAxis, rad2).toShape()
        cFace2 = Part.makeFace(circle2, "Part::FaceMakerSimple")

        shell = Part.makeShell((cFace1, hyperFace, cFace2))
        hyperboloid = Part.makeSolid(shell)
    else:
        shape = hyperbola.toBSpline(0, parameter).toShape(0, parameter)
        hyperFace = shape.revolve(center, axis, 360)

        rad = EndPoint.dot(hyperbola.YAxis)
        hgt = EndPoint.dot(hyperbola.XAxis)
        cc = center + hyperbola.XAxis * hgt
        circle = Part.Circle(cc, -hyperbola.XAxis, rad).toShape()
        cFace = Part.makeFace(circle, "Part::FaceMakerSimple")

        shell = Part.makeShell((cFace, hyperFace))
        hyper1 = Part.makeSolid(shell)
        hyper2 = hyper1.rotated(center, hyperbola.YAxis, 180)
        hyperboloid = Part.makeCompound((hyper1, hyper2))

    return hyperboloid


def makeHyperbolicCylinder(center, radii, rAxes, axis, length):

    S11 = center + rAxes[1] * radii[1]  # major axis
    S12 = center + rAxes[0] * radii[0]  # minor axis
    S21 = center - rAxes[1] * radii[1]  # major axis
    S22 = center - rAxes[0] * radii[0]  # minor axis

    hyperbola1 = Part.Hyperbola(S11, S12, center)
    hyperbola2 = Part.Hyperbola(S21, S22, center)
    d = axis * length

    Y = length
    X = radii[1] * math.sqrt((Y / radii[0]) ** 2 + 1)
    point = center + X * rAxes[1] + Y * rAxes[0]  # point in taken as length is always counted on minor axis
    parameter = abs(hyperbola1.parameter(point))

    shape1 = hyperbola1.toBSpline(-parameter, parameter).toShape(-parameter, parameter)
    shape2 = hyperbola2.toBSpline(-parameter, parameter).toShape(-parameter, parameter)
    surf1 = shape1.extrude(d)
    surf2 = shape2.extrude(d)

    return Part.makeCompound((surf1, surf2))


def makeEllipticCylinder(center, radii, rAxes, axis, length):

    S1 = center + rAxes[1] * radii[1]  # major axis
    S2 = center + rAxes[0] * radii[0]  # minor axis
    d = axis * length

    ellipse = Part.Ellipse(S1, S2, center)
    ellipse2 = Part.Ellipse(S1 + d, S2 + d, center + d)

    shape = ellipse.toBSpline().toShape()
    shape2 = ellipse2.toBSpline().toShape()
    shell = Part.makeLoft((shape, shape2), True)
    return Part.makeSolid(shell)


def makeEllipsoid(center, radii, rAxes, axis):

    S1 = center + rAxes[1] * radii[1]  # major axis
    S2 = center + rAxes[0] * radii[0]  # minor axis

    ellipse = Part.Ellipse(S1, S2, center)

    if axis.add(-rAxes[0]).Length < 1e-5:
        shape = ellipse.toBSpline().toShape()
        shell = shape.revolve(center, axis, 180)
    else:
        shape = ellipse.toBSpline(0, math.pi).toShape(0, math.pi)
        shell = shape.revolve(center, axis, 360)
    return Part.makeSolid(shell)


def makeEllipticTorus(R, RZ, RX, center, ZAxis):

    rMaj = RZ
    rMin = RX
    XAxis = ortoVect(ZAxis)

    majorAxis = ZAxis
    minorAxis = XAxis
    if rMaj < rMin:
        rMaj, rMin = rMin, rMaj
        majorAxis, minorAxis = minorAxis, majorAxis

    eCenter = center + R * XAxis
    S1 = eCenter + majorAxis * rMaj  # major axis
    S2 = eCenter + minorAxis * rMin  # minor axis

    ellipse = Part.Ellipse(S1, S2, eCenter)
    if abs(R) < RX:  # degenerated Torus
        pz = RZ * math.sqrt(1 - (R / RX) ** 2)
        pz1 = center - pz * ZAxis
        pz2 = center + pz * ZAxis

        p1 = ellipse.parameter(pz1)
        p2 = ellipse.parameter(pz2)
        if p2 < p1:
            p2 += 2 * math.pi
        shape = ellipse.toBSpline(p1, p2).toShape(p1, p2)  # revolution around Major axis
        rev = shape.revolve(center, ZAxis, 360)
    else:
        shape = ellipse.toBSpline().toShape()  # revolution around Minor axis
        rev = shape.revolve(center, ZAxis, 360)
    shell = Part.makeShell((rev,))
    return Part.makeSolid(shell)


def makeParaboloid(center, axis, focal, length):

    parabola = Part.Parabola()
    parabola.Center = center
    parabola.Axis = ortoVect(axis)
    parabola.XAxis = axis
    parabola.Focal = focal

    R = math.sqrt(4 * focal * length)
    point = center + length * parabola.XAxis + R * parabola.YAxis
    parameter = abs(parabola.parameter(point))

    shape = parabola.toBSpline(0, parameter).toShape(0, parameter)
    paraFace = shape.revolve(center, axis, 360)

    cc = center + length * parabola.XAxis
    circle = Part.Circle(cc, -axis, R).toShape()
    cFace = Part.makeFace(circle, "Part::FaceMakerSimple")

    shell = Part.makeShell((cFace, paraFace))
    return Part.makeSolid(shell)


def makeEllipticCone(apex, axis, Ra, radii, rAxes, length):

    S1 = apex + rAxes[1] * radii[1] / Ra * length  # major axis
    S2 = apex + rAxes[0] * radii[0] / Ra * length  # minor axis
    d = axis * length

    point = Part.Point(apex).toShape()
    ellipse = Part.Ellipse(S1 + d, S2 + d, apex + d)

    shape = ellipse.toBSpline().toShape()
    shell = Part.makeLoft((point, shape), True)
    return Part.makeSolid(shell)


def ortoVect(v):
    vmax = 0
    vOrto = None
    if abs(v.x) > vmax:
        vOrto = (0, 1, 0)
        vmax = abs(v.x)
    if abs(v.y) > vmax:
        vOrto = (0, 0, 1)
        vmax = abs(v.y)
    if abs(v.z) > vmax:
        vOrto = (1, 0, 0)
        vmax = abs(v.z)

    if vOrto is None:
        return None

    vOrto = v.cross(FreeCAD.Vector(vOrto))
    vOrto.normalize()
    return vOrto



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import re
import sys

sys.path.insert(0, os.path.abspath("../../src"))

# -- Project information -----------------------------------------------------

project = "GEOUNED"
copyright = "2024, UNED"

# The full version, including alpha/beta/rc tags
import geouned

version = geouned.__version__
release = geouned.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_autodoc_typehints",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.
# this theme supports versions https://github.com/pydata/pydata-sphinx-theme
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# TODO add logo
# html_favicon = "favicon.ico"

# Version match must match the 'version' key in version_switcher.json
pattern = re.compile(r"^[0-9]+\.[0-9]+")
version_match = pattern.search(version)
if version_match:
    version_match = version_match.group()
elif "dev" in version:
    version_match = "dev"
else:
    version_match = version

html_theme_options = {
    "github_url": "https://github.com/GEOUNED-org/GEOUNED",
    "switcher": {
        "json_url": "https://raw.githubusercontent.com/GEOUNED-org/GEOUNED/dev/docs/version_switcher.json",
        "version_match": version_match,
    },
    "nav_title": "GEOUNED",
    "navbar_start": ["version-switcher", "navbar-icon-links"],
}



import math
import os
import re
import xml.etree.ElementTree as ET

import FreeCAD
import numpy as np
from numpy import linalg as LA

from .Objects import CadCell, Cone, Cylinder, Plane, Sphere, Torus
from .XMLParser import get_cards


class XmlInput:
    def __init__(self, name):
        if not os.path.isfile(name):
            raise FileNotFoundError(f"File {name} does not exist")

        tree = ET.parse(name)
        root = tree.getroot()

        self.__inputcards__ = list(get_cards(root))
        return

    def GetFilteredCells(self, Surfaces, config):
        levels, contLevels, Universes = self.GetLevelStructure()

        FilteredCells = {}

        Ustart = config["Ustart"]
        subUniverses = getSubUniverses(Ustart, Universes)
        subUniverses.add(Ustart)

        if config["levelMax"] == "all":
            levelMax = len(levels)
        else:
            levelMax = config["levelMax"] + 1

        levelUniverse = set()
        for lev in range(0, levelMax):
            for U in levels[lev]:
                levelUniverse.add(U)
        subUniverses = subUniverses.intersection(levelUniverse)

        for U in list(Universes.keys()):
            if U not in subUniverses:
                del Universes[U]

        for U in Universes.keys():
            FilteredCells[U] = selectCells(Universes[U], config)
            processSurfaces(FilteredCells[U], Surfaces)

        # change the surface name in surface dict
        newSurfaces = {}
        for k in Surfaces.keys():
            newkey = Surfaces[k].id
            newSurfaces[newkey] = Surfaces[k]

        for U, universe in FilteredCells.items():

            # set cell as CAD cell Object
            for cname, c in universe.items():
                # print(cname,c.geom.str)
                universe[cname] = CadCell(c)

        return levels, FilteredCells, newSurfaces

    def GetLevelStructure(self):
        containers = []
        Universe_dict = {}

        for c in self.__inputcards__:
            if c.type != "cell":
                continue
            if c.U not in Universe_dict.keys():
                Universe_dict[c.U] = {}
            Universe_dict[c.U].update({c.name: c})

            if c.FILL:
                containers.append(c)

        currentLevel = [0]
        nextLevel = []
        contLevel = {0: [(0, 0)]}
        univLevel = {0: {0}}
        level = 0

        while True:
            level += 1
            contLevel[level] = []
            univLevel[level] = set()
            for c in reversed(containers):
                if c.U in currentLevel:
                    c.Level = level
                    nextLevel.append(c.FILL)
                    contLevel[level].append((c.U, c.name))
                    univLevel[level].add(c.FILL)

                    containers.remove(c)

            if nextLevel == []:
                break
            currentLevel = nextLevel
            nextLevel = []

        return univLevel, contLevel, Universe_dict

    def GetCells(self, U=None, Fill=None):
        cell_cards = {}
        for c in self.__inputcards__:
            if c.type != "cell":
                continue
            U_cell = c.U
            Fill_cell = c.FILL
            if U is None and Fill is None:
                cell_cards[c.name] = c
            elif U_cell == U and U is not None:
                cell_cards[c.name] = c
            elif Fill_cell == Fill and Fill is not None:
                cell_cards[c.name] = c

        return cell_cards

    def GetSurfaces(self, scale=1.0):
        surf_cards = {}
        number = 1
        for c in self.__inputcards__:
            if c.type != "surface":
                continue
            surf_cards[c.name] = (c.stype, c.scoefs, number)
            number += 1

        # return surface as surface Objects type
        return Get_primitive_surfaces(surf_cards, scale)


def selectCells(cellList, config):
    selected = {}
    # options are 'all' material
    if config["mat"][0] == "all":
        if config["cell"][0] == "all":
            selected = cellList
        elif config["cell"][0] == "exclude":
            for name, c in cellList.items():
                if name not in config["cell"][1]:
                    selected[name] = c
        elif config["cell"][0] == "include":
            for name, c in cellList.items():
                if name in config["cell"][1]:
                    selected[name] = c

    # options are 'exclude' material
    elif config["mat"][0] == "exclude":
        if config["cell"][0] == "all":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT not in config["mat"][1]:
                        selected[name] = c
                else:
                    selected[name] = c  # Fill cell are not tested against material number
        elif config["cell"][0] == "exclude":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT not in config["mat"][1]:
                        if name not in config["cell"][1]:
                            selected[name] = c
                else:
                    if name not in config["cell"][1]:
                        selected[name] = c  # Fill cell are not tested against material number
        elif config["cell"][0] == "include":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT not in config["mat"][1]:
                        if name in config["cell"][1]:
                            selected[name] = c
                else:
                    if name in config["cell"][1]:
                        selected[name] = c  # Fill cell are not tested against material number

    # options are 'include' material
    elif config["mat"][0] == "include":
        if config["cell"][0] == "all":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT in config["mat"][1]:
                        selected[name] = c
                else:
                    selected[name] = c  # Fill cell are not tested against material number
        elif config["cell"][0] == "exclude":
            for c in cellList:
                if c.FILL is None:
                    if c.MAT in config["mat"][1]:
                        if name not in config["cell"][1]:
                            selected[name] = c
                else:
                    if name not in config["cell"][1]:
                        selected[name] = c  # Fill cell are not tested against material number
        elif config["cell"][0] == "include":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT in config["mat"][1]:
                        if name in config["cell"][1]:
                            selected[name] = c
                else:
                    if name in config["cell"][1]:
                        selected[name] = c  # Fill cell are not tested against material number

    # remove complementary in cell of the universe
    # for cname,c in selected.items() :
    #   c.geom = remove_hash(cellList,cname)

    if not selected:
        raise ValueError("No cells selected. Check input or selection criteria in config file.")

    return selected


def processSurfaces(UCells, Surfaces):
    number = re.compile(r"\#?\s*\d+")

    for cname, c in UCells.items():
        pos = 0
        while True:
            m = number.search(c.geom.str, pos)
            if not m:
                break
            if "#" in m.group():
                pos = m.end()
                continue
            surf = int(m.group())
            if surf == 0:
                print(c.name)
                print(m)
                print(c.geom.str)
            pos = c.geom.replace(surf, Surfaces[surf].id, pos)


def getSubUniverses(Ustart, Universes):
    Uid = set()
    for c in Universes[Ustart].values():
        if c.FILL:
            Uid.add(c.FILL)

    AllU = Uid.copy()
    for U in Uid:
        AllU = AllU.union(getSubUniverses(U, Universes))

    return AllU


# traduce mcnp surface definition for Solid_Cell class
#  planes:
#     Stype = 'plane'
#     params = [ax,ay,az,d]
#
#  spheres:
#     Stype = 'shpere'
#     params = [cx,cy,cz,R]
#
#  cylinders:
#     Stype = 'cylinder'
#     params = [[px,py,pz],[vx,vy,vz],R]
#
#  cones:
#     Stype = 'cone'
#     params = [[px,py,pz],[vx,vy,vz],t,sht]
#
#  torus:
#     Stype = 'torus'
#     params = [[px,py,pz],[vx,vy,vz],ra,r]


# Return a diccionary with the corresponding surface Object
def Get_primitive_surfaces(mcnp_surfaces, scale=10.0):

    X_vec = FreeCAD.Vector(1.0, 0.0, 0.0)
    Y_vec = FreeCAD.Vector(0.0, 1.0, 0.0)
    Z_vec = FreeCAD.Vector(0.0, 0.0, 1.0)

    surfaces = {}
    for Sid in mcnp_surfaces.keys():
        MCNPtype = mcnp_surfaces[Sid][0]
        MCNPparams = mcnp_surfaces[Sid][1]
        number = mcnp_surfaces[Sid][2]

        params = []
        Stype = None
        if MCNPtype in ("plane", "x-plane", "y-plane", "z-plane"):
            Stype = "plane"
            if MCNPtype == "plane":
                normal = FreeCAD.Vector(MCNPparams[0:3])
                params = (normal, MCNPparams[3] * scale)
            elif MCNPtype == "x-plane":
                params = (X_vec, MCNPparams[0] * scale)
            elif MCNPtype == "y-plane":
                params = (Y_vec, MCNPparams[0] * scale)
            elif MCNPtype == "z-plane":
                params = (Z_vec, MCNPparams[0] * scale)

        elif MCNPtype == "sphere":
            Stype = "sphere"
            params = (FreeCAD.Vector(MCNPparams[0:3]) * scale, MCNPparams[3] * scale)

        elif MCNPtype in ("x-cylinder", "y-cylinder", "z-cylinder"):
            R = MCNPparams[2]
            x1 = MCNPparams[0]
            x2 = MCNPparams[1]
            Stype = "cylinder"

            if MCNPtype == "x-cylinder":
                v = X_vec
                p = FreeCAD.Vector(0.0, x1, x2)
            elif MCNPtype == "y-cylinder":
                v = Y_vec
                p = FreeCAD.Vector(x1, 0.0, x2)
            elif MCNPtype == "z-cylinder":
                v = Z_vec
                p = FreeCAD.Vector(x1, x2, 0.0)

            if scale != 1.0:
                p = p.multiply(scale)
                R *= scale

            params = (p, v, R)

        elif MCNPtype in ("x-cone", "y-cone", "z-cone"):
            Stype = "cone"
            x1 = MCNPparams[0]
            x2 = MCNPparams[1]
            x3 = MCNPparams[2]
            p = FreeCAD.Vector(x1, x2, x3)
            t2 = MCNPparams[3]
            t = math.sqrt(t2)
            dblsht = True

            if MCNPtype == "x-cone":
                v = X_vec
            elif MCNPtype == "y-cone":
                v = Y_vec
            elif MCNPtype == "z-cone":
                v = Z_vec

            p = p.multiply(scale)
            params = (p, v, t, dblsht)

        elif MCNPtype in ["x-torus", "y-torus", "z-torus"]:
            Stype = "torus"
            p = FreeCAD.Vector(MCNPparams[0:3])
            Ra, r1, r2 = MCNPparams[3:6]

            if MCNPtype == "x-torus":
                v = X_vec
            elif MCNPtype == "y-torus":
                v = Y_vec
            elif MCNPtype == "z-torus":
                v = Z_vec

            if scale != 1.0:
                Ra *= scale
                r1 *= scale
                r2 *= scale
                p = p.multiply(scale)

            params = (p, v, Ra, r1, r2)

        elif MCNPtype == "quadric":
            Qparams = tuple(MCNPparams[0:10])
            Stype, quadric = gq2cyl(Qparams)

            if Stype == "cylinder":
                p = FreeCAD.Vector(quadric[0:3])
                v = FreeCAD.Vector(quadric[3:6])
                R = quadric[6]
                if scale != 1.0:
                    R *= scale
                    p = p.multiply(scale)

                params = (p, v, R)

            elif Stype == "cone":
                p = FreeCAD.Vector(quadric[0:3])
                v = FreeCAD.Vector(quadric[3:6])
                t = quadric[6]
                dblsht = quadric[7]
                if scale != 1.0:
                    p = p.multiply(scale)

                params = (p, v, t, dblsht)

            else:
                print(Stype)
                params = None
        #                get_quadric_surface(params)

        if Stype == "plane":
            surfaces[Sid] = Plane(number, params)
        elif Stype == "sphere":
            surfaces[Sid] = Sphere(number, params)
        elif Stype == "cylinder":
            surfaces[Sid] = Cylinder(number, params)
        elif Stype == "cone":
            surfaces[Sid] = Cone(number, params)
        elif Stype == "torus":
            surfaces[Sid] = Torus(number, params)
        else:
            print("Undefined", Sid)
            print(MCNPtype, number, MCNPparams)

    return surfaces


def gq2cyl(x):
    # Conversion de GQ a Cyl
    # Ax2+By2+Cz2+Dxy+Eyz+Fxz+Gx+Hy+Jz+K=0
    # x.T*M*x + b.T*x + K = 0
    minWTol = 5.0e-2
    minRTol = 1.0e-3
    # minRTol=3.e-1
    # lx = np.array(x)
    tp = ""
    M = np.array(
        [
            [x[0], x[3] / 2, x[5] / 2],
            [x[3] / 2, x[1], x[4] / 2],
            [x[5] / 2, x[4] / 2, x[2]],
        ]
    )
    w, P = LA.eigh(M)
    sw = np.sort(w)
    aw = np.abs(w)
    asw = np.sort(aw)
    # Test for cylinder (least abs value is much less than others)
    if asw[0] < minWTol * asw[1]:
        tp = "cylinder"
        rv = [0.0] * 7  # X0,Y0,Z0, VX, VY, VZ, R
        iaxis = np.where(aw == asw[0])[0][0]
        otherAxes = ((iaxis + 1) % 3, (iaxis + 2) % 3)
        if abs(w[otherAxes[0]] - w[otherAxes[1]]) > minRTol * asw[2]:
            tp = "not found - ellipsoid cylinder"
            rv = [0]
            return tp, rv
        # Vector de desplazamiento
        # x0 = -0.5*Pt*D-1*P*b pero ojo que un lambda es cero
        # P es la matriz de valores propios
        b = np.array(x[6:9])
        Pb = np.matmul(P, b)
        for i in otherAxes:
            Pb[i] /= w[i]
        x0 = -0.5 * np.matmul(P.T, Pb)
        k = -0.5 * np.matmul(x0, b) - x[9]
        # Resultados finales

        rv[0:3] = x0  # Punto del eje
        rv[3:6] = P[:, iaxis]  # Vector director
        rv[6] = np.sqrt(k / sw[1])  # Radio
    # Test for cone (incomplete, returns empty data list)
    elif np.sign(sw[0]) != np.sign(sw[2]):  # maybe cone
        tp = "cone"
        rv = [0.0] * 8  #  X0, Y0, Z0, VX, VY, VZ, tgAlpha, double sheet
        if np.sign(sw[0]) == np.sign(sw[1]):
            iaxis = np.where(w == sw[2])[0][0]
        else:
            iaxis = np.where(w == sw[0])[0][0]
        otherAxes = ((iaxis + 1) % 3, (iaxis + 2) % 3)
        if abs(w[otherAxes[0]] - w[otherAxes[1]]) > minRTol * asw[2]:
            tp = "not found - ellipsoid cone/hyperboloid"
            rv = [0]
            return tp, rv
        # Displacement vector ( x0 = -0.5*M^-1*b = -0.5*P.T*D^-1*P*b
        b = np.array(x[6:9])
        x0 = -0.5 * np.matmul(P, np.matmul(P.T, b) / w)
        k = x0.T @ M @ x0 - x[9]
        if np.abs(k * w[iaxis]) > minRTol * minRTol * asw[2]:

            # tp = 'not found - hyperboloid'
            # rv = [0]
            # force cone surface
            print("Force cone surface")
            tp = "cone"
            rv[0:3] = x0  # vertex point
            rv[3:6] = P[:, iaxis]  # axis direction
            rv[6] = np.sqrt(-w[otherAxes[0]] / w[iaxis])  # semiangle tangent
            rv[7] = True  # here always double sheet cones
            return tp, rv
        # return value
        rv[0:3] = x0  # vertex point
        rv[3:6] = P[:, iaxis]  # axis direction
        rv[6] = np.sqrt(-w[otherAxes[0]] / w[iaxis])  # semiangle tangent
        rv[7] = True  # here always double sheet cones
    else:
        tp = "not found - unknown"
        rv = [0]
    return tp, rv



#
#   Conversion to MCNP v0.0
#   Only one solid and planar surfaces
#
import logging
import math

import FreeCAD

from .boolean_function import BoolSequence
from .functions import GeounedSurface, split_bop

BoolVals = (None, True, False)

logger = logging.getLogger("general_logger")


class CTelement:
    def __init__(self, val=None, S1=None, S2=None):
        self.diagonal = False
        self.S1 = S1
        self.S2 = S2
        if val is None:
            self.val = None
        else:
            if type(val) is int:
                self.diagonal = True
                self.type = None
            else:
                self.type = val.count(0)
            self.val = val

    def get_transpose(self):
        if self.diagonal:
            return self.val
        return CTelement((self.val[0], self.val[3], self.val[2], self.val[1]), self.S2, self.S1)

    def get_dependence(self):
        if self.diagonal:
            if self.val == 0:
                return True, False
            elif self.val == -1:
                return "Null", False
            else:
                return True, "Null"

        Ones = sum(self.val)
        if Ones == 4:
            return None, None
        elif Ones == 3:
            ind = self.val.index(0)
            if ind == 0:
                return False, None
            elif ind == 1:
                return True, None
            elif ind == 2:
                return None, True
            else:
                return None, False
        elif Ones == 2:
            ind1 = self.val.index(0)
            ind2 = self.val.index(0, ind1 + 1)
            if ind1 == 0 and ind2 == 1:
                return "Null", None
            elif ind1 == 2 and ind2 == 3:
                return None, "Null"
            elif ind1 == 0 and ind2 == 3:
                return False, False
            else:
                return True, True
        elif Ones == 1:
            ind = self.val.index(1)
            if ind == 0:
                return True, "Null"
            elif ind == 1:
                return False, "Null"
            elif ind == 2:
                return "Null", False
            else:
                return "Null", True
        else:
            return "Null", "Null"


class ConstraintTable(dict):

    def __init__(self):
        self.diagonal = None

    def __str__(self):

        varName = list(self.keys())

        # Constraint Table only diagonal terms
        if len(self[varName[0]]) == 1:
            line = ""
            for name in varName:
                element = self[name][name]
                line += f" {name:4d} : {element.val}\n"
            return line

        outstr = "  "
        for name in varName:
            outstr = outstr + f" {name:3d}"
        outstr = outstr + "\n"

        for name1 in varName:
            line = f" {name1:3d} "
            linenot = f"~{name1:3d} "
            for name2 in varName:
                elmt = self[name1][name2]
                if elmt.diagonal:
                    line += f" {elmt.val:>2d} "
                    linenot += "    "
                else:
                    line += f" {elmt.val[0]}{elmt.val[1]} "
                    linenot += f" {elmt.val[3]}{elmt.val[2]} "
            outstr += line + "\n"
            outstr += linenot + "\n"
        return outstr

    def add_element(self, k1, k2, val):

        if k1 in self.keys():
            self[k1][k2] = val
        else:
            self[k1] = {k2: val}

    def fill_missing_elements(self):
        keys = list(self.keys())
        missing = []
        for i, k1 in enumerate(keys):
            for k2 in keys[i + 1 :]:
                if k2 in self[k1].keys():
                    elmt = self[k1][k2]
                    self[k2][k1] = elmt.get_transpose()
                else:
                    missing.append((k1, k2))

        for k1, k2 in missing:
            diag1 = self[k1][k1]
            diag2 = self[k2][k2]
            new = combine_diag_elements(diag1, diag2)
            new.S1 = k1
            new.S2 = k2
            self[k1][k2] = new
            self[k2][k1] = new.get_transpose()

    def get_out_surfaces(self):
        out = []
        for k in self.keys():
            if self[k][k].val != 0:
                out.append(k)
        return out

    def get_constraint_set(self, valname):
        trueSet = {}
        falseSet = {}
        TNull = False
        FNull = False
        for k in self.keys():
            TValue, FValue = self[valname][k].get_dependence()
            if TValue == "Null":
                TNull = True
                TValue = None
            if FValue == "Null":
                FNull = True
                FValue = None
            if TValue is not None:
                trueSet[k] = TValue
            if FValue is not None:
                falseSet[k] = FValue

        if TNull:
            trueSet = None
        if FNull:
            falseSet = None
        return trueSet, falseSet

    def solid_in_box(self, Seq):  #  Sequence of the cell
        surfs = tuple(Seq.get_surfaces_numbers())
        if self.diagonal:
            seqValues = dict()
            for s in surfs:
                val = BoolVals[self[s][s].val]
                if val is not None:
                    seqValues[s] = val
            # if evaluate   None  : Box intersection Cell != 0      Part of the cell in the box
            #               True  : Box intersection Cell == Box    Cell cover the full region of the box. Void cell doesn't exist
            #               False : Box intersection Cell == 0      Cell out of the box
            res = Seq.evaluate(seqValues)
            return res if type(res) is bool else None

        else:
            trueSet, falseSet = self.get_constraint_set(surfs[0])
            if trueSet is not None:
                trueVal = Seq.evaluate(trueSet)
                trueVal = trueVal if type(trueVal) is bool else None
                if trueVal is None:
                    return None
                if falseSet is not None:
                    falseVal = Seq.evaluate(falseSet)
                    falseVal = falseVal if type(falseVal) is bool else None
                    if falseVal is None:
                        return None  # Part of the cell in the box
                    if trueVal == falseVal:
                        return trueVal  # True Cover full cell, False not in the box
                    else:
                        return None  # Part of the cell in the box
                else:
                    return trueVal
            elif falseSet is not None:
                falseVal = Seq.evaluate(falseSet)
                return falseVal if type(falseVal) is bool else None
            else:
                logger.info("Bad trouble surfaces is on none side of the box!!")
                return False


def combine_diag_elements(d1, d2):

    if d1.val == 0 and d2.val == 0:
        return CTelement((1, 1, 1, 1))
    elif d1.val == 1 and d2.val == 0:
        return CTelement((1, 1, 0, 0))
    elif d1.val == -1 and d2.val == 0:
        return CTelement((0, 0, 1, 1))
    elif d1.val == 0 and d2.val == 1:
        return CTelement((1, 0, 0, 1))
    elif d1.val == 0 and d2.val == -1:
        return CTelement((0, 1, 1, 0))
    elif d1.val == 1 and d2.val == 1:
        return CTelement((1, 0, 0, 0))
    elif d1.val == 1 and d2.val == -1:
        return CTelement((0, 1, 0, 0))
    elif d1.val == -1 and d2.val == 1:
        return CTelement((0, 0, 0, 1))
    elif d1.val == -1 and d2.val == -1:
        return CTelement((0, 0, 1, 0))


def build_c_table_from_solids(Box, SurfInfo, simplification_mode, options):

    if type(SurfInfo) is dict:
        surfaces = SurfInfo
        surfaceList = tuple(surfaces.keys())
    elif type(SurfInfo) is tuple:
        surfaceList, surfaces = SurfInfo
    else:
        surfaces = SurfInfo.Surfaces
        surfaceList = SurfInfo.surfaceList

    if type(surfaces[surfaceList[0]]) is GeounedSurface:
        for s in surfaceList:
            ss = surfaces[s]
            ss.__boundBox__ = Box.BoundBox
            ss.build_surface()
    else:
        for s in surfaceList:
            surfaces[s].buildShape(Box.BoundBox)

    CTable = ConstraintTable()
    if simplification_mode == "diag":
        CTable.diagonal = True
    else:
        CTable.diagonal = False

    for i, s1 in enumerate(surfaceList):
        res, splitRegions = split_solid_fast(Box, surfaces[s1], True, options)
        # res,splitRegions = split_solid_fast(Box,Surfaces.get_surface(s1),True)

        CTable.add_element(s1, s1, CTelement(res, s1, s1))
        if simplification_mode == "diag":
            continue
        if splitRegions is None:
            continue  # loop, no region to be split by s2
        for j, s2 in enumerate(surfaceList[i + 1 :]):
            posS1, negS1 = splitRegions

            pos0 = None
            for solid in posS1:

                pos = split_solid_fast(solid, surfaces[s2], False, options)

                # pos = split_solid_fast(solid,Surfaces.get_surface(s2),False)
                if pos == (1, 1):
                    break  # s2 intersect S1 Region
                if pos0 is None:
                    pos0 = pos
                else:
                    if pos != pos0:  # s1 regions are on both side of s2
                        pos = (1, 1)
                        break

            neg0 = None
            for solid in negS1:
                # neg = split_solid_fast(solid,Surfaces.get_surface(s2),False)
                neg = split_solid_fast(solid, surfaces[s2], False, options)
                if neg == (1, 1):
                    break  # s2 intersect S1 Region
                if neg0 is None:
                    neg0 = neg
                else:
                    if neg != neg0:  # s1 regions are on both side of s2
                        neg = (1, 1)
                        break

            val = (pos[0], pos[1], neg[1], neg[0])
            CTable.add_element(s1, s2, CTelement(val, s1, s2))

    # if some surfaces don't cross the box some elements in Constraint table are not filled
    if simplification_mode != "diag":
        CTable.fill_missing_elements()
    return CTable


def remove_extra_surfaces(CellSeq, CTable):
    # checking is make on solid cell definition to be removed from void cell
    outSurfaces = set(CTable.get_out_surfaces())
    newDef = BoolSequence(operator="OR")

    # Loop over all compound solids of the metaSolid

    if CellSeq.level == 0 and len(CellSeq.elements) == 1:
        return CellSeq

    if CellSeq.operator == "AND":
        newSeq = BoolSequence(operator="AND")
        newSeq.append(CellSeq.copy())
        CellSeq.assign(newSeq)

    for subCell in CellSeq.elements:
        nullcell = False
        subCell.check()
        if type(subCell.elements) is bool:
            chk = not subCell.elements
        else:
            chk = None

        if chk is False:  # the cell doesn't exist
            nullcell = True
        elif chk is True:  # the cell describe the full universe
            newDef.elements = True
            newDef.level = -1
            return newDef

        # if subcell has finite volume check it intersection with the box
        if not nullcell:
            res = CTable.solid_in_box(subCell)
            if res is None:
                # subcell intersect the box
                # get the surfaces of the solids out of the box
                # get reduced definition

                # if subcell lev!= 0 remove surface operation is not valid
                if subCell.level == 0:
                    removeSurf = outSurfaces & subCell.get_surfaces_numbers()
                    for s in removeSurf:
                        val = True if CTable[s][s].val > 0 else False
                        subCell.substitute(s, val)

                if type(subCell.elements) is bool:
                    if subCell.elements is False:  #  cell does not intersect void box
                        continue
                    else:  # cell cover fully void box
                        newDef.elements = True
                        newDef.level = -1
                        return newDef
                else:
                    newDef.append(subCell)

            elif res is True:
                # subcell cover the full box region Void cell doesn't exist
                newDef.elements = True
                newDef.level = -1
                return newDef

    newDef.clean()

    return newDef


def split_solid_fast(solid, surf, box, options):

    if box:
        if surf.shape:
            comsolid = split_bop(solid, [surf.shape], options.splitTolerance, options)
        else:
            return check_sign(solid, surf), None

        if len(comsolid.Solids) <= 1:
            return check_sign(solid, surf), None
        # sgn = check_sign(solid,surf)   # if "box" and single object => the box is not split, surface s1 out of the box.
        # if sgn == 1 :                 # The sign is the side of surface s1 where the box is located
        #    # return ((1,0),(0,0)),None  # return the diagonal element of the Constraint Table for s1
        #    return ((1,0),(0,0)),None  # return the diagonal element of the Constraint Table for s1
        # else:
        #    return ((0,0),(0,1)),None

        else:
            posSol = []
            negSol = []
            for s in comsolid.Solids:
                sgn = check_sign(s, surf)
                if sgn == 1:
                    posSol.append(s)
                else:
                    negSol.append(s)
            return 0, (
                posSol,
                negSol,
            )  # return the diagonal element of the Constraint Table for s1, and solids to be split by s2
            # return ((1,0),(0,1)), (posSol,negSol) # return the diagonal element of the Constraint Table for s1, and solids to be split by s2

    else:
        # "not box" => return the position of the +/- region of s1 (the solid) with respect s2
        if surf.shape:
            dist = solid.distToShape(surf.shape)[0]
        else:
            dist = 1.0
        if dist > 1e-6:  # face doesn't intersect solid
            sgn = check_sign(solid, surf)
            if sgn == 1:
                return (
                    1,
                    0,
                )  # values of s2 and -s2   "0" means this region doesn't exist
            else:
                return (0, 1)
        else:
            return (1, 1)  # values of s2 and -s2


# find one point inside a solid (region)
def point_inside(solid):

    point = solid.CenterOfMass
    if solid.isInside(point, 0.0, False):
        return point

    cut_line = 32
    cut_box = 2

    v1 = solid.Vertexes[0].Point
    for vi in range(len(solid.Vertexes) - 1, 0, -1):
        v2 = solid.Vertexes[vi].Point
        dv = (v2 - v1) * 0.5

        n = 1
        while True:
            for i in range(n):
                point = v1 + dv * (1 + 0.5 * i)
                if solid.isInside(point, 0.0, False):
                    return point
            n = n * 2
            dv = dv * 0.5
            if n > cut_line:
                break

    BBox = solid.optimalBoundingBox(False)
    box = [BBox.XMin, BBox.XMax, BBox.YMin, BBox.YMax, BBox.ZMin, BBox.ZMax]

    boxes, centers = divide_box(box)
    n = 0

    while True:
        for p in centers:
            pp = FreeCAD.Vector(p[0], p[1], p[2])
            if solid.isInside(pp, 0.0, False):
                return pp

        subbox = []
        centers = []
        for b in boxes:
            btab, ctab = divide_box(b)
            subbox.extend(btab)
            centers.extend(ctab)
        boxes = subbox
        n = n + 1

        if n == cut_box:
            break

    return point_from_surface(solid)


def point_from_surface(solid):

    for face in solid.Faces:
        parameters = face.ParameterRange
        u = (parameters[0] + parameters[1]) / 2
        v = (parameters[2] + parameters[3]) / 2
        pface = face.valueAt(u, v)
        normal = face.normalAt(u, v)

        d = 10
        pp = pface + d * normal
        while d > 1e-8:
            if solid.isInside(pp, 0.0, False):
                return pp
            d *= 0.5
            pp = pface + d * normal

        normal = -normal
        d = 10
        pp = pface + d * normal
        while d > 1e-8:
            if solid.isInside(pp, 0.0, False):
                return pp
            d *= 0.5
            pp = pface + d * normal

    logger.info(f"Solid not found in bounding Box (Volume : {solid.Volume})")
    return None


# divide a box into 8 smaller boxes
def divide_box(Box):
    xmid = (Box[1] + Box[0]) * 0.5
    ymid = (Box[3] + Box[2]) * 0.5
    zmid = (Box[5] + Box[4]) * 0.5

    b1 = (Box[0], xmid, Box[2], ymid, Box[4], zmid)
    p1 = (0.5 * (Box[0] + xmid), 0.5 * (Box[2] + ymid), 0.5 * (Box[4] + zmid))

    b2 = (xmid, Box[1], Box[2], ymid, Box[4], zmid)
    p2 = (0.5 * (xmid + Box[1]), 0.5 * (Box[2] + ymid), 0.5 * (Box[4] + zmid))

    b3 = (Box[0], xmid, ymid, Box[3], Box[4], zmid)
    p3 = (0.5 * (Box[0] + xmid), 0.5 * (ymid + Box[3]), 0.5 * (Box[4] + zmid))

    b4 = (xmid, Box[1], ymid, Box[3], Box[4], zmid)
    p4 = (0.5 * (xmid + Box[1]), 0.5 * (ymid + Box[3]), 0.5 * (Box[4] + zmid))

    b5 = (Box[0], xmid, Box[2], ymid, zmid, Box[5])
    p5 = (0.5 * (Box[0] + xmid), 0.5 * (Box[2] + ymid), 0.5 * (zmid + Box[5]))

    b6 = (xmid, Box[1], Box[2], ymid, zmid, Box[5])
    p6 = (0.5 * (xmid + Box[1]), 0.5 * (Box[2] + ymid), 0.5 * (zmid + Box[5]))

    b7 = (Box[0], xmid, ymid, Box[3], zmid, Box[5])
    p7 = (0.5 * (Box[0] + xmid), 0.5 * (ymid + Box[3]), 0.5 * (zmid + Box[5]))

    b8 = (xmid, Box[1], ymid, Box[3], zmid, Box[5])
    p8 = (0.5 * (xmid + Box[1]), 0.5 * (ymid + Box[3]), 0.5 * (zmid + Box[5]))

    return (b1, b2, b3, b4, b5, b6, b7, b8), (p1, p2, p3, p4, p5, p6, p7, p8)


def check_sign(solid, surf):

    point = point_inside(solid)

    if surf.Type == "Plane":
        r = point - surf.Surf.Position
        if surf.Surf.Axis.dot(r) > 0:
            return 1
        else:
            return -1

    elif surf.Type == "Cylinder":
        r = point - surf.Surf.Center
        L2 = r.Length * r.Length
        z = surf.Surf.Axis.dot(r)
        z2 = z * z
        R2 = surf.Surf.Radius * surf.Surf.Radius
        if L2 - z2 > R2:
            return 1
        else:
            return -1

    elif surf.Type == "Sphere":
        r = point - surf.Surf.Center
        if r.Length > surf.Surf.Radius:
            return 1
        else:
            return -1

    elif surf.Type == "Cone":
        r = point - surf.Surf.Apex
        r.normalize()
        z = round(surf.Surf.Axis.dot(r), 15)
        alpha = math.acos(z)

        if alpha > surf.Surf.SemiAngle:
            return 1
        else:
            return -1

    elif surf.Type == "Torus":
        r = point - surf.Surf.Center
        h = r.dot(surf.Surf.Axis)
        rho = r - h * surf.Surf.Axis

        rp = math.sqrt((rho.Length - surf.Surf.MajorRadius) ** 2 + h**2)
        if rp > surf.Surf.MinorRadius:
            return 1
        else:
            return -1



import configparser

from .options import Options


def setOptions(optionFile):

    setting = {
        "fileIn": "",
        "fileOut": "",
        "outBox": tuple(),
        "inFormat": "mcnp",
        "UStart": 0,
        "levelMax": "all",
        "cell": ("all",),
        "mat": ("exclude", (0,)),
    }

    sectionNames = ("Setting", "Levels", "Cells", "Materials", "Options")

    config = configparser.ConfigParser()
    config.optionxform = str
    config.read(optionFile)

    fileData = False
    for section in config.sections():
        if section not in sectionNames:
            raise ValueError(f"{section} bad section name")

        if section == "Setting":
            fileIn, fileOut, outBox, inFormat = getSetting(config)

            i = 0
            if fileIn is not None:
                setting["fileIn"] = fileIn
                i += 1

            if fileOut is not None:
                setting["fileOut"] = fileOut
                i += 1

            if outBox is not None:
                setting["outBox"] = outBox
                i += 1

            if inFormat is not None:
                setting["inFormat"] = inFormat

            if i == 3:
                fileData = True

        elif section == "Levels":
            UStart, levMax = getLevel(config)
            if UStart is not None:
                setting["UStart"] = UStart
            if levMax is not None:
                setting["levelMax"] = levMax

        elif section in ("Cells", "Materials"):
            CMRange = getRange(section, config)
            if CMRange is None:
                continue
            if section == "Cells":
                setting["cell"] = CMRange
            else:
                setting["mat"] = CMRange

        elif section == "Options":
            setSecOptions(config)

    if not fileData:
        raise ValueError("missing input data in [Setting] section")

    return setting


def getSetting(config):

    fileIn, fileOut, outBox, inFormat = None, None, None, None
    for key in config["Setting"].keys():

        if key in ["inputFile", "mcnpFile"]:
            fileIn = config.get("Setting", key)
        elif key == "CADFile":
            fileOut = config.get("Setting", key)
        elif key == "outBox":
            outBox = getBoxData(config.get("Setting", key))
        elif key == "inFormat":
            inFormat = config.get("Setting", key)
        else:
            print(f"{key} bad keyword. Ignored")

    return fileIn, fileOut, outBox, inFormat


def getLevel(config):

    UStart, levMax = None, None
    for key in config["Levels"].keys():
        if key == "levelMax":
            levMax = config.get("Levels", key)
            if levMax.lower() == "all":
                levMax = "all"
            elif levMax.isnumeric():
                levMax = int(UStart)
            else:
                print("bad value for levMax")
        elif key == "UStart":
            UStart = config.getint("Levels", key)
        else:
            print(f"{key} bad keyword. Ignored")

    return UStart, levMax


def getRange(section, config):
    rType = None
    rawRange = None
    for key in config[section].keys():
        if key == "rangeType":
            rType = config.get(section, key)
        elif key == "range":
            rawRange = config.get(section, key)
        else:
            print(f"{key} bad keyword. Ignored")

    if rType is None:
        return None
    elif rType.lower() == "all":
        return ("all",)

    elif rType.lower() == "exclude" or rType.lower() == "include":
        ctRange = getRangeData(rawRange)
        if ctRange is None:
            print(f"bad range in section {section}")
            return None
        else:
            return (rType.lower(), ctRange)
    else:
        return None


def getBoxData(string):
    data = tuple(map(float, string.split()))
    if len(data) != 6:
        raise ValueError("bad Outbox value number")

    elif (data[0] > data[1]) or (data[2] > data[3]) or (data[4] > data[5]):
        raise ValueError("bad Outbox boundaries")

    else:
        return (
            10 * data[0],
            10 * data[2],
            10 * data[4],
            10 * data[1],
            10 * data[3],
            10 * data[5],
        )


def getRangeData(rawData):
    if rawData is None:
        return None

    intervals = rawData.split(",")
    rangeValues = set()

    for i in intervals:
        if i.strip().isnumeric():
            rangeValues.add(int(i))
        else:
            bounds = i.split(":")
            if len(bounds) != 2:
                print("bad range definition")
                return None
            else:
                if bounds[0].strip().isnumeric() and bounds[1].strip().isnumeric():
                    i1, i2 = map(int, bounds)
                    if i1 > i2:
                        print("bad range definition")
                        return None
                    else:
                        rangeValues.update(range(i1, i2 + 1))
                else:
                    print("bad range definition")
                    return None

    rangeValues = list(rangeValues)
    rangeValues.sort()
    return tuple(rangeValues)


def setSecOptions(config):

    for key in config["Options"].keys():
        if key == "splitTolerance":
            tolerance = config.getfloat("Options", key)
            setattr(Options, key, tolerance)
        else:
            print(f"{key} bad keyword. Ignored")

    return


def rangeGenerator(intervals):
    for i in intervals:
        if type(i) is tuple:
            for v in range(i[0], i[1] + 1):
                yield v
        else:
            yield i



#
# Script to obtain the Qform of a Cylinder
#
#
#

import math

import FreeCAD


def rotation_matrix(u, v):
    """Definition of the rotation matrix for two vectors"""

    # defintion of the axis of rotation

    Axis = u.cross(v).normalize()

    u.normalize()
    v.normalize()

    cose = u.dot(v)
    seno = u.cross(v).Length

    R = FreeCAD.Matrix()

    # cose=math.cos(angle)
    # seno=math.sin(angle)
    onecos = 1.0 - cose

    # 1st row
    R.A11 = cose + Axis.x**2 * onecos
    R.A12 = Axis.x * Axis.y * onecos - Axis.z * seno
    R.A13 = Axis.x * Axis.z * onecos + Axis.y * seno

    # 2nd row
    R.A21 = Axis.x * Axis.y * onecos + Axis.z * seno
    R.A22 = cose + Axis.y**2 * onecos
    R.A23 = Axis.y * Axis.z * onecos - Axis.x * seno

    # 3rd row
    R.A31 = Axis.z * Axis.x * onecos - Axis.y * seno
    R.A32 = Axis.z * Axis.y * onecos + Axis.x * seno
    R.A33 = cose + Axis.z**2 * onecos

    return R


def q_form_cyl(Axis, Pos, rad):

    R = rotation_matrix(FreeCAD.Vector(1, 0, 0), Axis)
    R.transpose()
    Pos2 = R.multiply(Pos).negative()

    A = R.A21**2 + R.A31**2
    B = R.A22**2 + R.A32**2
    C = R.A23**2 + R.A33**2

    D = 2.0 * (R.A21 * R.A22 + R.A31 * R.A32)
    E = 2.0 * (R.A22 * R.A23 + R.A32 * R.A33)
    F = 2.0 * (R.A23 * R.A21 + R.A33 * R.A31)

    G = 2.0 * (Pos2.y * R.A21 + Pos2.z * R.A31)
    H = 2.0 * (Pos2.y * R.A22 + Pos2.z * R.A32)
    J = 2.0 * (Pos2.y * R.A23 + Pos2.z * R.A33)

    K = Pos2.y**2 + Pos2.z**2 - rad**2

    return (A, B, C, D, E, F, G, H, J, K)


def q_form_cone(Axis, Pos, tan):

    R = rotation_matrix(FreeCAD.Vector(1, 0, 0), Axis)
    R.transpose()
    Pos2 = R.multiply(Pos).negative()

    A = R.A21**2 + R.A31**2 - (tan * R.A11) ** 2
    B = R.A22**2 + R.A32**2 - (tan * R.A12) ** 2
    C = R.A23**2 + R.A33**2 - (tan * R.A13) ** 2

    D = 2.0 * (R.A21 * R.A22 + R.A31 * R.A32 - tan**2 * R.A11 * R.A12)
    E = 2.0 * (R.A22 * R.A23 + R.A32 * R.A33 - tan**2 * R.A12 * R.A13)
    F = 2.0 * (R.A23 * R.A21 + R.A33 * R.A31 - tan**2 * R.A13 * R.A11)

    G = 2.0 * (Pos2.y * R.A21 + Pos2.z * R.A31 - tan**2 * Pos2.x * R.A11)
    H = 2.0 * (Pos2.y * R.A22 + Pos2.z * R.A32 - tan**2 * Pos2.x * R.A12)
    J = 2.0 * (Pos2.y * R.A23 + Pos2.z * R.A33 - tan**2 * Pos2.x * R.A13)

    K = Pos2.y**2 + Pos2.z**2 - (tan * Pos2.x) ** 2

    return (A, B, C, D, E, F, G, H, J, K)



import logging
import re

import FreeCAD

from ..utils.functions import GeounedSolid

logger = logging.getLogger("general_logger")


def get_label(label, options):
    """Deleting the last word of the string if this is a number
    Only if the option delLastNumber is True"""
    if not options.delLastNumber:
        return label
    wrd = label.split()
    try:
        new_label = " ".join(wrd[:-1])
        return new_label
    except:
        return label


def getCommentTree(obj, options):
    recursive_list = []
    c_obj = obj
    while c_obj.InList:
        label = get_label(c_obj.InList[0].Label, options)
        recursive_list.append(label)
        c_obj = c_obj.InList[0]

    comment = ""
    for label in reversed(recursive_list):
        comment += "/" + label
    return comment


def joinEnvelopes(meta_list):
    join_list = []
    old_id = -1
    new_list = []
    for i, m in enumerate(meta_list):
        if m.CellType != "envelope":
            continue
        if m.EnclosureID != old_id:
            join_list.append(new_list)
            new_list = [i]
            old_id = m.EnclosureID
        else:
            new_list.append(i)

    join_list.append(new_list)
    del join_list[0]

    for id_list in join_list:
        fuse_meta_obj(meta_list, id_list[0], id_list[-1] + 1)


def fuse_meta_obj(meta_list, init, end):
    if end - init == 1:
        meta_list[init].__id__ = init + 1
        return
    solids = []
    for m in meta_list[init:end]:
        solids.extend(m.Solids)

    new_meta = GeounedSolid(init + 1, solids)
    new_meta.EnclosureID = meta_list[init].EnclosureID
    new_meta.ParentEnclosureID = meta_list[init].ParentEnclosureID
    new_meta.IsEnclosure = meta_list[init].IsEnclosure
    new_meta.CellType = meta_list[init].CellType

    del meta_list[init:end]
    meta_list.insert(init, new_meta)


# Paco mod
# TODO check if this function is actually used in to code
def check_cad_file_material(material, mdict):
    """Function to check that if a component in the CAD file has a material, it exist in the material file"""
    templist = [elem for elem in set(material) if elem not in mdict]
    if templist:
        raise ValueError(
            "At least one material in the CAD model is not present in the material file\n"
            f"List of not present materials:{templist}\n"
            "Code STOPS\n"
        )
    return


# Paco mod
def set_enclosure_solid_list(meta_list):
    """Function to define DataList list."""
    return [elem for elem in meta_list if elem.IsEnclosure]


def remove_enclosure(meta_list):
    """This function removes Enclosure solid and actualises __id__ attributes to take into account the removal from meta_list list of nested enclosures originally loaded from CAD model."""

    # update id of solids
    len_meta_list = len(meta_list)
    i_count = 0
    for i, m in enumerate(meta_list):
        if m.IsEnclosure:
            i_count += 1
        else:
            meta_list[i].__id__ -= i_count

    # remove Enclosure solids from metaList
    for i in range(len_meta_list - 1, -1, -1):
        if meta_list[i].IsEnclosure:
            meta_list.pop(i)


def set_enclosure_levels(enclosure_list):
    nested_list = [[0], []]
    nested_enclosure = [[]]
    parent_level = 0
    temp_list = enclosure_list[:]

    while len(temp_list) > 0:
        remove = []
        for i, encl in enumerate(temp_list):
            if encl.ParentEnclosureID in nested_list[parent_level]:
                nested_list[parent_level + 1].append(encl.EnclosureID)
                nested_enclosure[parent_level].append(encl)
                remove.append(i)

        nested_list.append([])
        nested_enclosure.append([])
        parent_level += 1
        remove.reverse()
        for i in remove:
            del temp_list[i]

    nested_enclosure_list = []
    for Level in nested_enclosure:
        temp = []
        for i, encl in enumerate(Level):
            temp.append((encl.ParentEnclosureID, i))
        temp.sort()

        grouped = []
        for t in temp:
            grouped.append(Level[t[1]])
        nested_enclosure_list.append(grouped)

    for i, Level in enumerate(nested_enclosure_list[0:-1]):
        for encl in Level:
            child_list = []
            for child in nested_enclosure_list[i + 1]:
                if child.ParentEnclosureID == encl.EnclosureID:
                    child_list.append(child)
            encl.SonEnclosures = child_list[:]

    return nested_enclosure_list


def check_enclosure(freecad_doc, enclosure_list):

    stop = False
    # check all enclosure labels have an associated solid
    temp_list = []
    for elem in freecad_doc.Objects:
        if elem.TypeId == "Part::Feature":
            if elem.Shape.Solids:
                if elem.InList:
                    templabel = re.search(
                        "enclosure(?P<encl>[0-9]+)_(?P<parent>[0-9]+)_",
                        elem.InList[0].Label,
                    )
                    if templabel is not None:
                        if elem.TypeId == "Part::Feature" and len(elem.Shape.Solids) == 0:
                            temp_list.append(elem)

    if temp_list:
        stop = True
        logger.info("One or more nested enclosure labels in CAD solid tree view/structure tree do not have any CAD solid.")
        logger.info("Each nested enclosure must have only one solid. Code STOPS.")
        logger.info("List of problematic nested enclosure labels:")

        for elem in temp_list:
            logger.info(elem.EnclosureID)

    # check enclosure Labels don't make loops

    # 1) check at leat one 0 is in Parent enclosure label
    # 2) check 0 is not in child enclosure label
    # 3) check child enclosure label is not repeated

    sid_list = []
    pid_set = set()
    repeated_id = set()

    for encl in enclosure_list:
        pid_set.add(encl.ParentEnclosureID)
        if encl.EnclosureID in sid_list:
            repeated_id.add(encl.EnclosureID)
        else:
            sid_list.append(encl.EnclosureID)

    if 0 in sid_list:
        stop = True
        logger.info('"0" cannot be label on child Enclosure')
    if 0 not in pid_set:
        stop = True
        logger.info('"0" should parent label of most external enclosure(s)')
    if repeated_id:
        stop = True
        logger.info("Child label cannot be repeated.\nRepeated labels :")
        for lab in repeated_id:
            logger.info(lab)

    # this stop should not be move to the end of routine
    # if previous point not satisfied point 4) may lead to infinite loop
    if stop:
        raise ValueError("Exiting to avoid infinite loop")

    # 4) look for explicit loops
    encl_dict = dict()
    for encl in enclosure_list:
        if encl.ParentEnclosureID in encl_dict.keys():
            encl_dict[encl.ParentEnclosureID].append(encl.EnclosureID)
        else:
            encl_dict[encl.ParentEnclosureID] = [encl.EnclosureID]

    parent = [0]
    cont = True
    while cont:
        cont = False
        next_parent = []
        for p in parent:
            if p in encl_dict.keys():
                cont = True
                next_parent.extend(encl_dict[p])
                del encl_dict[p]
        parent = next_parent

    if encl_dict.keys():
        logger.info("Following enclosure produce loop")
        for p in encl_dict.keys():
            for s in encl_dict[p]:
                logger.info(f"{s}_{p}")
        raise ValueError("GEOUNED.LoadFunctions.check_enclosure failed")

    # check enclosures solids are correctly nested and don't overlap each other
    nested_levels = set_enclosure_levels(enclosure_list)

    overlap = []
    encl_tree = [[0]]
    for level in nested_levels:
        encl_tree = update_tree(encl_tree, level)
        for encl in level:
            same_parent = dict()
            if encl.ParentEnclosureID in same_parent.keys():
                same_parent[encl.ParentEnclosureID].append(encl.EnclosureID)
            else:
                same_parent[encl.ParentEnclosureID] = [encl.EnclosureID]
        for encl in same_parent.values():
            overlap.extend(check_overlap(encl))

    not_embedded = []
    for chain in encl_tree:
        up = chain[0]
        for low in chain[1:]:
            inter = up.check_intersection(low.CADSolid)
            if inter != -2:
                not_embedded.append((low.EnclosureID, up.EnclosureID))

    if not_embedded:
        stop = True
        logger.info("Following enclosures are not fully embedded in Parent enclosure")
        for elemt in not_embedded:
            logger.info(f"{elemt[0]}_{elemt[1]}")

    if overlap:
        stop = True
        logger.info("Following enclosures overlapping ")
        for elemt in overlap:
            logger.info(f"{elemt[0]}_{elemt[1]}")

    if stop:
        raise ValueError("GEOUNED.LoadFunctions.check_enclosure failed")


def check_overlap(enclosures):
    overlap = []
    for i, enc1 in enumerate(enclosures):
        for enc2 in enclosures[i + 1 :]:
            inter = enc1.check_intersection(enc2.CADSolid)
            if inter != 1:
                overlap.append((enc1.EnclosureID, enc2.EnclosureID))
    return overlap


def update_tree(Tree, level):
    new_tree = []
    for encl in level:
        for lst in Tree:
            if lst[-1] == encl.ParentEnclosureID:
                new_tree.append(lst + [encl.EnclosureID])
                continue
    return new_tree


def set_doc_options():
    # set import step document options for FreeCAD version >0.18 compatible with FreeCAD0.18 opening options
    p0 = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Import")
    p0.SetBool("UseLinkGroup", False)
    p0.SetBool("ReduceObjects", False)
    p0.SetBool("ExpandCompound", False)

    p1 = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Import/hSTEP")
    p1.SetBool("UseLinkGroup", False)
    p1.SetBool("ReadShapeCompoundMode", True)


def check_index(docList, index, returnObject=False):
    subList = docList.ElementList
    for i in index[:-1]:
        if len(subList) > i:
            subList = subList[i]
            if subList.TypeId != "App::LinkGroup":
                return None
            subList = subList.ElementList
        else:
            return None

    if len(subList) > index[-1]:
        if subList[index[-1]].TypeId == "App::LinkGroup":
            return "Link"
        else:
            if returnObject:
                return subList[index[-1]]
            else:
                return "Feature"
    else:
        return None


def next_index(docList, lastIndex=None):
    if lastIndex is None:
        lastIndex = [0]
        while check_index(docList, lastIndex) != "Feature":
            lastIndex.append(0)
        else:
            return lastIndex

    tmp = lastIndex[:]
    move = True
    while True:
        if move:
            tmp[-1] = tmp[-1] + 1
        obj = check_index(docList, tmp)
        if obj == "Feature":
            return tmp
        elif obj == "Link":
            tmp.append(0)
            move = False
        else:
            tmp = tmp[0:-1]
            move = True
            if len(tmp) == 0:
                return None


# TODO check this function is used in the code
def solid_generator(doclist):
    last = None
    while True:
        last = next_index(doclist, last)
        if last is None:
            break
        yield check_index(doclist, last, True)



"""
The tests check the resulting CSG geometry in particle transport using OpenMC.
The tests assumes that the test_convert script has previously been run and
that xml files exist in the tests_outputs folder.
"""

import os
import sys
from pathlib import Path

try:
    import freecad  # importing conda package if present
except:
    pass
import openmc
import Part
import pytest
from FreeCAD import Import

openmc.config["cross_sections"] = Path("tests/cross_sections.xml").resolve()

path_to_cad = Path("testing/inputSTEP")
step_files = list(path_to_cad.rglob("*.stp")) + list(path_to_cad.rglob("*.step"))

# these step files all produced failures, most of them lose particles
# they are removed from the tests but an issue has be raised and a minimal
# example has been made in the issue where the fixing of these files can be
# tracked https://github.com/shimwell/GEOUNED/issues/11
step_files.remove(Path("testing/inputSTEP/large/SCDR.stp"))
step_files.remove(Path("testing/inputSTEP/placa.stp"))
step_files.remove(Path("testing/inputSTEP/placa2.stp"))
# required a few more particles needed but also losing particles
step_files.remove(Path("testing/inputSTEP/DoubleCylinder/placa3.step"))
step_files.remove(Path("testing/inputSTEP/DoubleCylinder/placa.stp"))
# this face2.stp crashes when loading the geometry.xml

# removing geometries that are particularly slow to convert from CI testing
# these geometries remain in the test suite for locally testing
if os.getenv("GITHUB_ACTIONS"):
    step_files.remove(Path("testing/inputSTEP/large/Triangle.stp"))


@pytest.mark.skipif(
    sys.platform in ["win32", "darwin"],
    reason="OpenMC doesn't install on Windows currently and is not well tested on Mac",
)
@pytest.mark.parametrize("input_step_file", step_files)
def test_transport(input_step_file):

    output_dir = Path("tests_outputs") / input_step_file.with_suffix("")
    output_dir.mkdir(parents=True, exist_ok=True)
    output_filename_stem = output_dir / input_step_file.stem
    openmc_xml_file = output_filename_stem.with_suffix(".xml")

    Import.insert(str(input_step_file), "converted-cad")
    result = Part.Shape()
    result.read(str(input_step_file))

    # getting the bounding box of the CAD so we can use it to make a source
    # over the whole geometry
    bb = result.BoundBox
    llx, lly, llz, urx, ury, urz = bb.XMin, bb.YMin, bb.ZMin, bb.XMax, bb.YMax, bb.ZMax
    # converting from mm to cm
    llx, lly, llz, urx, ury, urz = (
        llx / 10,
        lly / 10,
        llz / 10,
        urx / 10,
        ury / 10,
        urz / 10,
    )

    source = openmc.IndependentSource()
    source.space = openmc.stats.Box(lower_left=(llx, lly, llz), upper_right=(urx, ury, urz))
    source.energy = openmc.stats.Discrete([14e6], [1])

    materials = openmc.Materials()
    geometry = openmc.Geometry.from_xml(openmc_xml_file, materials)

    settings = openmc.Settings()
    settings.batches = 10
    settings.max_lost_particles = 1
    # number of particles is increased for local runs as the GitHub action
    # runner has two threads and a typical local computer has more.
    # Sometimes a lot of particles are needed to find small geometry errors
    if os.getenv("GITHUB_ACTIONS"):
        settings.particles = 100_000
    else:
        settings.particles = 1_000_000
    settings.run_mode = "fixed source"
    settings.source = source
    model = openmc.Model(geometry, materials, settings)

    model.run()



""" Starting beta stable version 0.9

 - new in version 0.9.1
   - read MCNP macrobodies : RPP, BOX, RCC, REC, TRC
   - read SQ surfaces
   - hyperboloid 1 or 2 sheets (revolution), paraboloid (revolution), elliptical and hyperbolic cylinder, 
     elliptical cone and elliptical torus surfaces have been added.

   Changes have not been fully checked, some bugs may still remains. Please contact the authors in case of failure.
"""

GEOReverse_Version = "0.9.1"
GEOReverse_ReleaseDate = "16/11/2023"
